# TODO: 
# backend/idaes_service/solver/custom/steam_header.py [DONE]
# backend/idaes_service/solver/custom/tests/test_steam_header.py [DONE]
# backend/idaes_service/solver/methods/adapter_library.py
# backend/idaes_factory/adapters/unit_models/header_adapter.py
# backend/idaes_factory/adapters/adapter_library.py
# backend/core/auxiliary/enums/unitOpData.py
# backend/flowsheetInternals/unitops/config/config_base.py
# backend/flowsheetInternals/unitops/config/objects/header_config.py
# frontend/src/data/objects.json

from pyomo.environ import Suffix, Var
from pyomo.common.config import ConfigBlock, ConfigValue
from pyomo.network import Arc
from pyomo.core.base.reference import Reference
import pyomo.environ as pyo

# Import IDAES cores
from idaes.core import (
    declare_process_block_class,
    UnitModelBlockData,
    useDefault,
)
from idaes.core.util.config import is_physical_parameter_block
import idaes.logger as idaeslog
from idaes.core.util.tables import create_stream_table_dataframe
from idaes.models.unit_models.mixer import Mixer
from idaes.models.unit_models.heater import Heater

# Set up logger
_log = idaeslog.getLogger(__name__)


@declare_process_block_class("Desuperheater")
class DesuperheaterData(UnitModelBlockData):
    """
    Desuperheater unit operation:
    Superheated steam + water -> Mixer -> Losses -> Saturated Vapour
    Desuperheater aims to remove steam superhear by direct contact with boiler feed water.
    Uses Sequential Decomposition for optimal initialization order.
    """

    CONFIG = UnitModelBlockData.CONFIG()
    CONFIG.declare(
        "property_package",
        ConfigValue(
            default=useDefault,
            domain=is_physical_parameter_block,
            description="Property package to use for control volume",
        ),
    )
    CONFIG.declare(
        "property_package_args",
        ConfigBlock(
            implicit=True,
            description="Arguments to use for constructing property packages",
        ),
    )


    def build(self):
        super().build()
        self.scaling_factor = Suffix(direction=Suffix.EXPORT)

        # Create internal units
        self.mixer = Mixer(
            property_package=self.config.property_package,
            property_package_args=self.config.property_package_args,
            num_inlets=2
        )
        self.losses = Heater(
            property_package=self.config.property_package,
            property_package_args=self.config.property_package_args,
            has_pressure_change=True,
        )

        # Build saturated vapour state block
        tmp_dict = dict(**self.config.property_package_args)
        tmp_dict["has_phase_equilibrium"] = True
        tmp_dict["defined_state"] = False

        self._properties_sat = self.config.property_package.build_state_block(
            self.flowsheet().time, 
            doc="saturated vapour properties at outlet", 
            **tmp_dict
        )

        self.unit_ops = [self.mixer, self.losses, self._properties_sat]

        # Create internal arcs
        self.condenser_to_subcooler = Arc(
            source=self.mixer.outlet,
            destination=self.losses.inlet,
        )

        # Expand arcs
        pyo.TransformationFactory("network.expand_arcs").apply_to(self)

        # Identify the exposed ports
        inlet_exposed_ls = [
            (self.mixer, "steam_inlet", "inlet_1"),
            (self.mixer, "water_inlet", "inlet_2"),
        ]
        outlet_exposed_ls = [
            (self.losses, "steam_outlet", "outlet"),
        ]

        self.inlet_list, self.inlet_blocks = self._construct_exposed_ports(inlet_exposed_ls)
        self.outlet_list, self.outlet_blocks = self._construct_exposed_ports(outlet_exposed_ls)
        
        # Declare additional variables and provide alias of existing ones to expose to the user        
        self.deltaT_supheat = Var(
            self.flowsheet().time, 
            domain=pyo.NonNegativeReals,
            # initialize=0.0, 
            units=pyo.units.K, 
            doc="The degree of steam superheat remaining."
        )
        self.water_temperature = Reference(
            self.water_inlet_state[:].temperature
        )        
        self.water_flow_mol = Reference(
            self.water_inlet_state[:].flow_mol
        )
        self.water_flow_mass = Reference(
            self.water_inlet_state[:].flow_mass
        )
        self.heat_duty = Reference(
            self.losses.heat_duty[:]
        )
        self.deltaP = Reference(
            self.losses.deltaP[:]
        ) 

        # Declare internal variables and references for internal use
        self._T_sat = Reference(
            self._properties_sat[:].temperature
        )

        # Additional bounds and constraints
        self._additional_constraints()
               
    def _additional_constraints(self):
        """
        Additional constraints.
        """
        
        """
        1. Pressure equality for saturation state block
        """
        @self.Constraint(
            self.flowsheet().time, doc="Pressure equality for saturation state block"
        )
        def outlet_pressure_eql(b, t):
            return (
                b._properties_sat[t].pressure == b.losses.control_volume.properties_out[t].pressure
            )

        """
        2. Vapour quality for saturation state block
        """
        @self.Constraint(
            self.flowsheet().time, doc="Pressure equality for saturation state block"
        )
        def vapour_quality_eq(b, t):
            return (
                b._properties_sat[t].vapor_frac == 0.5
            )
        
        """
        3. Mixer pressure inlet equality for steam and water 
        """
        @self.Constraint(
                self.flowsheet().time, doc="Mixer pressure inlet equality for steam and water"
        )
        def mixer_inlet_pressure_eql(b, t):
            return (
                b.steam_inlet_state[t].pressure == b.water_inlet_state[t].pressure
            )
                
        """
        4. Superheat temperature balance
        """
        @self.Constraint(self.flowsheet().time, doc="Temperature superheat balance")
        def superheat_temperature_eqn(b, t):
            return (
                b.deltaT_supheat[t] + 0.001 == b.steam_outlet_state[t].temperature - b._T_sat[t]
            )
                
    def _construct_exposed_ports(self, exposed_ls):
        """
        exposed_ls: list of [(unit, public_name, internal_basename)]
        """
        names, states = [], []

        for unit, public_name, internal_basename in exposed_ls:
            # 1) Locate the state block on the child unit
            # Try <basename>_state (Mixer & many IDAES units)
            state = getattr(unit, f"{internal_basename}_state", None)

            # Fall back to control_volume properties for CV-based units (e.g., Heater)
            if state is None and hasattr(unit, "control_volume"):
                cv = unit.control_volume
                if internal_basename.startswith("inlet") and hasattr(cv, "properties_in"):
                    state = cv.properties_in
                elif internal_basename.startswith("outlet") and hasattr(cv, "properties_out"):
                    state = cv.properties_out

            if state is None:
                raise AttributeError(
                    f"{unit.name} has no state for '{internal_basename}'. "
                    f"Tried '{internal_basename}_state' and control_volume properties."
                )

            # 2) Expose the state on this wrapper
            #    (a) keep a convenient Reference to the time-indexed state block
            setattr(self, f"{public_name}_state", Reference(state[:]))

            #    (b) create a public Port mapped to that state
            #        NOTE: UnitModelBlock.add_port builds a Port with the vars
            #        defined by the property package’s port members.
            self.add_port(name=public_name, block=state)

            names.append(public_name)
            states.append(state)

        return names, states

    def calculate_scaling_factors(self):
        super().calculate_scaling_factors()
        for o in self.unit_ops:
            if hasattr(o, "calculate_scaling_factors"):
                o.calculate_scaling_factors()
                
    def _get_stream_table_contents(self, time_point=0):
        """
        Create stream table showing all inlets and outlets
        """
        io_dict = {}
        
        for inlet_name in self.inlet_list:
            io_dict[inlet_name] = getattr(self, inlet_name)
        
        for outlet_name in self.outlet_list:
            io_dict[outlet_name] = getattr(self, outlet_name)
            
        return create_stream_table_dataframe(io_dict, time_point=time_point)
     
    def _copy_port_state(self, src_port, snk_port, t):
        for name in ("flow_mol", "pressure", "enth_mol", "temperature", "enth_mass", "vapor_frac"):
            if hasattr(src_port, name) and hasattr(snk_port, name):
                src = getattr(src_port, name)
                snk = getattr(snk_port, name)
                if (t in src) and (t in snk) and pyo.is_variable_type(snk[t]):
                    try:
                        snk[t].set_value(pyo.value(src[t]))
                    except Exception:
                        pass  

    def _hold_state_vars(self, sb):
        held = {}
        if hasattr(sb, "define_state_vars"):
            for k, v in sb.define_state_vars().items():
                if pyo.is_variable_type(v) and not v.fixed:
                    v.fix()
                    held[k] = False
                else:
                    held[k] = True
        return held

    def _release_state_vars(self, sb, held):
        if hasattr(sb, "define_state_vars"):
            for k, was_fixed in held.items():
                v = sb.define_state_vars()[k]
                if was_fixed is False and pyo.is_variable_type(v):
                    v.unfix()

    def initialize_build(self, outlvl=idaeslog.NOTSET, **kwargs):
        """
        Initialize the Desuperheater without using state_args for sub-units.

        Steps:
        1) Initialize Mixer (no state_args; it reads from connected inlet ports).
        2) Initialize Heater with temporary ΔP=0 and Q=0, then restore.
        3) Seed and (if supported) initialize saturated-vapor state block.
        4) Seed deltaT_supheat.
        """
        init_log = idaeslog.getInitLogger(self.name, outlvl, tag="unit")
        solve_log = idaeslog.getSolveLogger(self.name, outlvl, tag="unit")

        t0 = list(self.flowsheet().time)[0]

        # 1) Mixer
        init_log.info("Initializing Mixer (no state_args)...")
        try:
            # Inlet states should already be provided via self.steam_inlet / self.water_inlet
            # and are wired to mixer.inlet_1 / mixer.inlet_2 via the exposed ports.
            self.water_flow_mol[t0].fix(0.0)
            self.water_flow_mol[t0].unfix()
            self.mixer.inlet_2.pressure[t0].fix(
                pyo.value(
                    self.mixer.inlet_1.pressure[t0]
                )
            )
            self.mixer.inlet_2.pressure[t0].unfix()
            self.mixer.initialize(outlvl=outlvl)
        except Exception as err:
            init_log.error(f"Mixer initialization failed: {err}")
            raise

        # 2) Heater (losses)
        init_log.info("Initializing Heater (losses) with temporary ΔP=0 and Q=0...")
        self._copy_port_state(
            src_port=self.mixer.outlet,
            snk_port=self.losses.inlet,
            t=t0
        )
        held = self._hold_state_vars(self.losses.control_volume.properties_in[t0])
        try:
            self.losses.initialize(outlvl=outlvl)
        finally:
            self._release_state_vars(self.losses.control_volume.properties_in[t0], held)

        # 3) Saturated-vapor state block seeding
        init_log.info("Seeding saturated-vapor state block...")
        try:
            # Set seed values directly
            self._properties_sat[t0].pressure.set_value(
                pyo.value(self.losses.outlet.pressure[t0])
            )
            self._properties_sat[t0].enth_mol.set_value(
                pyo.value(self.losses.outlet.enth_mol[t0])
            )

            # If the block exposes its own initialize(outlvl=...), call it
            if hasattr(self._properties_sat[t0], "initialize"):
                # Many property packages implement initialize on the *indexed* block,
                # i.e. call on the parent indexed component:
                try:
                    self._properties_sat.initialize(outlvl=outlvl)
                except TypeError:
                    # Some implementations only support per-timepoint initialize
                    self._properties_sat[t0].initialize(outlvl=outlvl)

        except Exception as err:
            init_log.error(f"Saturation state initialization failed: {err}")
            raise

        init_log.info("Desuperheater initialization complete.")



import pytest
 
from pyomo.environ import (
    check_optimal_termination,
    ConcreteModel,
    Constraint,
    Expression,
    Param,
    units as pyunits,
    value,
    Var,
)
from pyomo.util.check_units import assert_units_consistent, assert_units_equivalent
 
from idaes.core import FlowsheetBlock
from idaes.models.unit_models.heat_exchanger_ntu import (
    HeatExchangerNTU as HXNTU,
    HXNTUInitializer,
)
 
from idaes.models.properties.modular_properties.base.generic_property import (
    GenericParameterBlock,
)
from idaes.models_extra.column_models.properties.MEA_solvent import (
    configuration as aqueous_mea,
)
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.core.solvers import get_solver
from idaes.core.util.testing import initialization_tester, PhysicalParameterTestBlock
from idaes.core.util.exceptions import ConfigurationError, InitializationError
from idaes.core.initialization import (
    BlockTriangularizationInitializer,
    InitializationStatus,
)
from idaes.core.util import DiagnosticsToolbox    
from idaes.models.properties import iapws95

def test_ntu_hx():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)
    m.fs.properties = iapws95.Iapws95ParameterBlock(
    phase_presentation=iapws95.PhaseType.MIX,
    state_vars=iapws95.StateVars.PH
    )
    import math
    
    m.fs.unit = HXNTU(
        hot_side={
            "property_package": m.fs.properties,
            "has_pressure_change": False,
        },
        cold_side={
            "property_package": m.fs.properties,
            "has_pressure_change": False,
        },
    )
    
    # Hot fluid
    m.fs.unit.hot_side_inlet.flow_mol[0].fix(60.54879)
    m.fs.unit.hot_side_inlet.enth_mol[0].fix(m.fs.properties.htpx(p=202650*pyunits.Pa, T=392.23*pyunits.K))
    #m.fs.unit.hot_side_inlet.temperature[0].fix(392.23)
    m.fs.unit.hot_side_inlet.pressure[0].fix(202650)
    
    
    # Cold fluid
    m.fs.unit.cold_side_inlet.flow_mol[0].fix(63.01910)
    m.fs.unit.cold_side_inlet.enth_mol[0].fix(m.fs.properties.htpx(p=202650*pyunits.Pa, T=326.36*pyunits.K))
    #m.fs.unit.cold_side_inlet.temperature[0].fix(326.36)
    m.fs.unit.cold_side_inlet.pressure[0].fix(202650)
    
    
    # Unit design variables
    m.fs.unit.area.fix(100)
    m.fs.unit.heat_transfer_coefficient.fix(200)
    e = 0.7
    m.fs.unit.effectiveness.fix(e)
    
    # m.fs.unit.hot_side.deltaP.fix(-2000)
    # m.fs.unit.cold_side.deltaP.fix(-2000)
    
    m.fs.unit.initialize()
    
    solver = get_solver()
    results = solver.solve(m, tee=True)
    
    m.fs.unit.report()
    print(value(m.fs.unit.NTU[0]))
    print(value(m.fs.unit.heat_duty[0]))
    print(value(m.fs.unit.Cmin[0]))
    print(value(m.fs.unit.Cmax[0]))
    
    if value(m.fs.unit.hot_side.properties_in[0].flow_mol * m.fs.unit.hot_side.properties_in[0].cp_mol) < value(m.fs.unit.cold_side.properties_in[0].flow_mol * m.fs.unit.cold_side.properties_in[0].cp_mol):
        Cmin = value(
            m.fs.unit.hot_side.properties_in[0].flow_mol
            * m.fs.unit.hot_side.properties_in[0].cp_mol
        )
        Cmax = value(
            m.fs.unit.cold_side.properties_in[0].flow_mol
            * m.fs.unit.cold_side.properties_in[0].cp_mol
        )
    else:
        Cmax = value(
            m.fs.unit.hot_side.properties_in[0].flow_mol
            * m.fs.unit.hot_side.properties_in[0].cp_mol
        )
        Cmin = value(
                m.fs.unit.hot_side.properties_in[0].flow_mol
                * m.fs.unit.hot_side.properties_in[0].cp_mol
            )
    
    NTU = value(m.fs.unit.heat_transfer_coefficient[0] * m.fs.unit.area / Cmin)
    Q_duty = value(Cmin * e * (m.fs.unit.hot_side.properties_in[0].temperature - m.fs.unit.cold_side.properties_in[0].temperature))
    # Cmin = 5
    # Cmax = 5
    #NTU = 18
    #Q_duty = 20
    assert math.isclose(value(m.fs.unit.Cmin[0]), Cmin, rel_tol=1e-5), (
        f"Cmin mismatch: model={value(m.fs.unit.Cmin[0])}, calculated={Cmin}"
    )
    
    assert math.isclose(value(m.fs.unit.Cmax[0]), Cmax, rel_tol=1e-5), (
        f"Cmax mismatch: model={value(m.fs.unit.Cmax[0])}, calculated={Cmax}"
    )
    assert math.isclose(value(m.fs.unit.NTU[0]), NTU, rel_tol=1e-5), (
        f"NTU mismatch: model={value(m.fs.unit.NTU[0])}, calculated={NTU}"
    )
    assert math.isclose(value(m.fs.unit.heat_duty[0]), Q_duty, rel_tol=1e-5), (
        f"Heat duty mismatch: model={value(m.fs.unit.heat_duty[0])}, calculated={Q_duty}"
    )


##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Example ideal parameter block for the VLE calucations for a
Benzene-Toluene-o-Xylene system.
"""


# Import Pyomo libraries
from pyomo.environ import (
    Constraint,
    Expression,
    log,
    NonNegativeReals,
    Var,
    Set,
    Param,
    sqrt,
    log10,
    units as pyunits,
)
from pyomo.util.calc_var_value import calculate_variable_from_constraint
from pyomo.common.config import ConfigValue

# Import IDAES cores
from idaes.core import (
    declare_process_block_class,
    MaterialFlowBasis,
    PhysicalParameterBlock,
    StateBlockData,
    StateBlock,
    MaterialBalanceType,
    EnergyBalanceType,
    Component,
    LiquidPhase,
    VaporPhase,
)
from idaes.core.util.constants import Constants as const
from idaes.core.util.initialization import fix_state_vars, solve_indexed_blocks
from idaes.core.initialization import InitializerBase
from idaes.core.util.misc import add_object_reference
from idaes.core.util.model_statistics import number_unfixed_variables
from idaes.core.util.misc import extract_data
from idaes.core.solvers import get_solver
import idaes.core.util.scaling as iscale
import idaes.logger as idaeslog

# Set up logger
_log = idaeslog.getLogger(__name__)


class HDAInitializer(InitializerBase):
    """
    Initializer for HDA Property package.

    """

    CONFIG = InitializerBase.CONFIG()
    CONFIG.declare(
        "solver",
        ConfigValue(default=None, domain=str, description="Initialization solver"),
    )
    CONFIG.declare(
        "solver_options",
        ConfigValue(default=None, description="Initialization solver options"),
    )

    def initialization_routine(self, blk):
        init_log = idaeslog.getInitLogger(
            blk.name, self.config.output_level, tag="properties"
        )
        solve_log = idaeslog.getSolveLogger(
            blk.name, self.config.output_level, tag="properties"
        )

        # Set solver
        solver = get_solver(self.config.solver, self.config.solver_options)

        # ---------------------------------------------------------------------
        # If present, initialize bubble and dew point calculations
        for k in blk.keys():
            if hasattr(blk[k], "eq_temperature_dew"):
                calculate_variable_from_constraint(
                    blk[k].temperature_dew, blk[k].eq_temperature_dew
                )

            if hasattr(blk[k], "eq_pressure_dew"):
                calculate_variable_from_constraint(
                    blk[k].pressure_dew, blk[k].eq_pressure_dew
                )

        init_log.info_high(
            "Initialization Step 1 - Dew and bubble points " "calculation completed."
        )

        # ---------------------------------------------------------------------
        # If flash, initialize T1 and Teq
        for k in blk.keys():
            if blk[k].config.has_phase_equilibrium and not blk[k].config.defined_state:
                blk[k]._t1.value = max(
                    blk[k].temperature.value, blk[k].temperature_bubble.value
                )
                blk[k]._teq.value = min(blk[k]._t1.value, blk[k].temperature_dew.value)

        init_log.info_high(
            "Initialization Step 2 - Equilibrium temperature " " calculation completed."
        )

        # ---------------------------------------------------------------------
        # Initialize flow rates and compositions
        free_vars = 0
        for k in blk.keys():
            free_vars += number_unfixed_variables(blk[k])
        if free_vars > 0:
            try:
                with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
                    res = solve_indexed_blocks(solver, [blk], tee=slc.tee)
            except:
                res = None
        else:
            res = None

        init_log.info("Initialization Complete")

        return res


@declare_process_block_class("HDAParameterBlock")
class HDAParameterData(PhysicalParameterBlock):
    CONFIG = PhysicalParameterBlock.CONFIG()

    def build(self):
        """
        Callable method for Block construction.
        """
        super(HDAParameterData, self).build()

        self._state_block_class = IdealStateBlock

        self.benzene = Component()
        self.toluene = Component()
        self.methane = Component()
        self.hydrogen = Component()

        self.Liq = LiquidPhase()
        self.Vap = VaporPhase()

        # List of components in each phase (optional)
        self.phase_comp = {"Liq": self.component_list, "Vap": self.component_list}

        # List of phase equilibrium index
        self.phase_equilibrium_idx = Set(initialize=[1, 2, 3, 4])

        self.phase_equilibrium_list = {
            1: ["benzene", ("Vap", "Liq")],
            2: ["toluene", ("Vap", "Liq")],
            3: ["hydrogen", ("Vap", "Liq")],
            4: ["methane", ("Vap", "Liq")],
        }

        # Thermodynamic reference state
        self.pressure_ref = Param(
            mutable=True, default=101325, units=pyunits.Pa, doc="Reference pressure"
        )
        self.temperature_ref = Param(
            mutable=True, default=298.15, units=pyunits.K, doc="Reference temperature"
        )

        # Source: The Properties of Gases and Liquids (1987)
        # 4th edition, Chemical Engineering Series - Robert C. Reid
        pressure_crit_data = {
            "benzene": 48.9e5,
            "toluene": 41e5,
            "hydrogen": 12.9e5,
            "methane": 46e5,
        }

        self.pressure_crit = Param(
            self.component_list,
            within=NonNegativeReals,
            mutable=True,
            units=pyunits.Pa,
            initialize=extract_data(pressure_crit_data),
            doc="Critical pressure",
        )

        # Source: The Properties of Gases and Liquids (1987)
        # 4th edition, Chemical Engineering Series - Robert C. Reid
        temperature_crit_data = {
            "benzene": 562.2,
            "toluene": 591.8,
            "hydrogen": 33.0,
            "methane": 190.4,
        }

        self.temperature_crit = Param(
            self.component_list,
            within=NonNegativeReals,
            mutable=True,
            units=pyunits.K,
            initialize=extract_data(temperature_crit_data),
            doc="Critical temperature",
        )

        # Source: The Properties of Gases and Liquids (1987)
        # 4th edition, Chemical Engineering Series - Robert C. Reid
        mw_comp_data = {
            "benzene": 78.1136e-3,
            "toluene": 92.1405e-3,
            "hydrogen": 2.016e-3,
            "methane": 16.043e-3,
        }

        self.mw_comp = Param(
            self.component_list,
            mutable=True,
            units=pyunits.kg / pyunits.mol,
            initialize=extract_data(mw_comp_data),
            doc="molecular weight",
        )

        # Constants for liquid densities
        # Source: Perry's Chemical Engineers Handbook
        #         - Robert H. Perry (Cp_liq)
        dens_liq_data = {
            ("benzene", "1"): 1.0162,
            ("benzene", "2"): 0.2655,
            ("benzene", "3"): 562.16,
            ("benzene", "4"): 0.28212,
            ("toluene", "1"): 0.8488,
            ("toluene", "2"): 0.26655,
            ("toluene", "3"): 591.8,
            ("toluene", "4"): 0.2878,
            ("hydrogen", "1"): 5.414,
            ("hydrogen", "2"): 0.34893,
            ("hydrogen", "3"): 33.19,
            ("hydrogen", "4"): 0.2706,
            ("methane", "1"): 2.9214,
            ("methane", "2"): 0.28976,
            ("methane", "3"): 190.56,
            ("methane", "4"): 0.28881,
        }

        self.dens_liq_param_1 = Param(
            self.component_list,
            mutable=True,
            initialize={c: v for (c, j), v in dens_liq_data.items() if j == "1"},
            doc="Parameter 1 to compute liquid densities",
            units=pyunits.kmol * pyunits.m**-3,
        )

        self.dens_liq_param_2 = Param(
            self.component_list,
            mutable=True,
            initialize={c: v for (c, j), v in dens_liq_data.items() if j == "2"},
            doc="Parameter 2 to compute liquid densities",
            units=pyunits.dimensionless,
        )

        self.dens_liq_param_3 = Param(
            self.component_list,
            mutable=True,
            initialize={c: v for (c, j), v in dens_liq_data.items() if j == "3"},
            doc="Parameter 3 to compute liquid densities",
            units=pyunits.K,
        )

        self.dens_liq_param_4 = Param(
            self.component_list,
            mutable=True,
            initialize={c: v for (c, j), v in dens_liq_data.items() if j == "4"},
            doc="Parameter 4 to compute liquid densities",
            units=pyunits.dimensionless,
        )

        # Boiling point at standard pressure
        # Source: Perry's Chemical Engineers Handbook
        #         - Robert H. Perry (Cp_liq)
        bp_data = {
            ("benzene"): 353.25,
            ("toluene"): 383.95,
            ("hydrogen"): 20.45,
            ("methane"): 111.75,
        }

        self.temperature_boil = Param(
            self.component_list,
            mutable=True,
            units=pyunits.K,
            initialize=extract_data(bp_data),
            doc="Pure component boiling points at standard pressure",
        )

        # Constants for specific heat capacity, enthalpy
        # Sources: The Properties of Gases and Liquids (1987)
        #         4th edition, Chemical Engineering Series - Robert C. Reid
        #         Perry's Chemical Engineers Handbook
        #         - Robert H. Perry (Cp_liq)
        cp_ig_data = {
            ("Liq", "benzene", "1"): 1.29e5,
            ("Liq", "benzene", "2"): -1.7e2,
            ("Liq", "benzene", "3"): 6.48e-1,
            ("Liq", "benzene", "4"): 0,
            ("Liq", "benzene", "5"): 0,
            ("Vap", "benzene", "1"): -3.392e1,
            ("Vap", "benzene", "2"): 4.739e-1,
            ("Vap", "benzene", "3"): -3.017e-4,
            ("Vap", "benzene", "4"): 7.130e-8,
            ("Vap", "benzene", "5"): 0,
            ("Liq", "toluene", "1"): 1.40e5,
            ("Liq", "toluene", "2"): -1.52e2,
            ("Liq", "toluene", "3"): 6.95e-1,
            ("Liq", "toluene", "4"): 0,
            ("Liq", "toluene", "5"): 0,
            ("Vap", "toluene", "1"): -2.435e1,
            ("Vap", "toluene", "2"): 5.125e-1,
            ("Vap", "toluene", "3"): -2.765e-4,
            ("Vap", "toluene", "4"): 4.911e-8,
            ("Vap", "toluene", "5"): 0,
            ("Liq", "hydrogen", "1"): 0,  # 6.6653e1,
            ("Liq", "hydrogen", "2"): 0,  # 6.7659e3,
            ("Liq", "hydrogen", "3"): 0,  # -1.2363e2,
            ("Liq", "hydrogen", "4"): 0,  # 4.7827e2, # Eqn 2
            ("Liq", "hydrogen", "5"): 0,
            ("Vap", "hydrogen", "1"): 2.714e1,
            ("Vap", "hydrogen", "2"): 9.274e-3,
            ("Vap", "hydrogen", "3"): -1.381e-5,
            ("Vap", "hydrogen", "4"): 7.645e-9,
            ("Vap", "hydrogen", "5"): 0,
            ("Liq", "methane", "1"): 0,  # 6.5708e1,
            ("Liq", "methane", "2"): 0,  # 3.8883e4,
            ("Liq", "methane", "3"): 0,  # -2.5795e2,
            ("Liq", "methane", "4"): 0,  # 6.1407e2, # Eqn 2
            ("Liq", "methane", "5"): 0,
            ("Vap", "methane", "1"): 1.925e1,
            ("Vap", "methane", "2"): 5.213e-2,
            ("Vap", "methane", "3"): 1.197e-5,
            ("Vap", "methane", "4"): -1.132e-8,
            ("Vap", "methane", "5"): 0,
        }

        self.cp_ig_1 = Param(
            self.phase_list,
            self.component_list,
            mutable=True,
            initialize={(p, c): v for (p, c, j), v in cp_ig_data.items() if j == "1"},
            doc="Parameter 1 to compute Cp_comp",
            units=pyunits.J / pyunits.mol / pyunits.K,
        )

        self.cp_ig_2 = Param(
            self.phase_list,
            self.component_list,
            mutable=True,
            initialize={(p, c): v for (p, c, j), v in cp_ig_data.items() if j == "2"},
            doc="Parameter 2 to compute Cp_comp",
            units=pyunits.J / pyunits.mol / pyunits.K**2,
        )

        self.cp_ig_3 = Param(
            self.phase_list,
            self.component_list,
            mutable=True,
            initialize={(p, c): v for (p, c, j), v in cp_ig_data.items() if j == "3"},
            doc="Parameter 3 to compute Cp_comp",
            units=pyunits.J / pyunits.mol / pyunits.K**3,
        )

        self.cp_ig_4 = Param(
            self.phase_list,
            self.component_list,
            mutable=True,
            initialize={(p, c): v for (p, c, j), v in cp_ig_data.items() if j == "4"},
            doc="Parameter 4 to compute Cp_comp",
            units=pyunits.J / pyunits.mol / pyunits.K**4,
        )

        self.cp_ig_5 = Param(
            self.phase_list,
            self.component_list,
            mutable=True,
            initialize={(p, c): v for (p, c, j), v in cp_ig_data.items() if j == "5"},
            doc="Parameter 5 to compute Cp_comp",
            units=pyunits.J / pyunits.mol / pyunits.K**5,
        )

        # Source: The Properties of Gases and Liquids (1987)
        # 4th edition, Chemical Engineering Series - Robert C. Reid
        # fitted to Antoine form
        # H2, Methane from NIST webbook
        pressure_sat_coeff_data = {
            ("benzene", "A"): 4.202,
            ("benzene", "B"): 1322,
            ("benzene", "C"): -38.56,
            ("toluene", "A"): 4.216,
            ("toluene", "B"): 1435,
            ("toluene", "C"): -43.33,
            ("hydrogen", "A"): 3.543,
            ("hydrogen", "B"): 99.40,
            ("hydrogen", "C"): 7.726,
            ("methane", "A"): 3.990,
            ("methane", "B"): 443.0,
            ("methane", "C"): -0.49,
        }

        self.pressure_sat_coeff_A = Param(
            self.component_list,
            mutable=True,
            initialize={
                c: v for (c, j), v in pressure_sat_coeff_data.items() if j == "A"
            },
            doc="Parameter A to compute saturated pressure",
            units=pyunits.dimensionless,
        )

        self.pressure_sat_coeff_B = Param(
            self.component_list,
            mutable=True,
            initialize={
                c: v for (c, j), v in pressure_sat_coeff_data.items() if j == "B"
            },
            doc="Parameter B to compute saturated pressure",
            units=pyunits.K,
        )

        self.pressure_sat_coeff_C = Param(
            self.component_list,
            mutable=True,
            initialize={
                c: v for (c, j), v in pressure_sat_coeff_data.items() if j == "C"
            },
            doc="Parameter C to compute saturated pressure",
            units=pyunits.K,
        )

        # Source: The Properties of Gases and Liquids (1987)
        # 4th edition, Chemical Engineering Series - Robert C. Reid
        dh_vap = {"benzene": 3.387e4, "toluene": 3.8262e4, "hydrogen": 0, "methane": 0}

        self.dh_vap = Param(
            self.component_list,
            mutable=True,
            units=pyunits.J / pyunits.mol,
            initialize=extract_data(dh_vap),
            doc="heat of vaporization",
        )

        # Set default scaling factors
        self.set_default_scaling("flow_mol", 1e3)
        self.set_default_scaling("flow_mol_phase_comp", 1e3)
        self.set_default_scaling("flow_mol_phase", 1e3)
        self.set_default_scaling("material_flow_terms", 1e3)
        self.set_default_scaling("enthalpy_flow_terms", 1e-2)
        self.set_default_scaling("mole_frac_comp", 1e1)
        self.set_default_scaling("temperature", 1e-2)
        self.set_default_scaling("temperature_dew", 1e-2)
        self.set_default_scaling("temperature_bubble", 1e-2)
        self.set_default_scaling("pressure", 1e-5)
        self.set_default_scaling("pressure_sat", 1e-5)
        self.set_default_scaling("pressure_dew", 1e-5)
        self.set_default_scaling("pressure_bubble", 1e-5)
        self.set_default_scaling("mole_frac_phase_comp", 1e1)
        self.set_default_scaling("enth_mol_phase", 1e-3, index="Liq")
        self.set_default_scaling("enth_mol_phase", 1e-4, index="Vap")
        self.set_default_scaling("enth_mol", 1e-3)
        self.set_default_scaling("entr_mol_phase", 1e-2)
        self.set_default_scaling("entr_mol", 1e-2)

    @classmethod
    def define_metadata(cls, obj):
        """Define properties supported and units."""
        obj.add_properties(
            {
                "flow_mol": {"method": None},
                "flow_mol_phase_comp": {"method": None},
                "mole_frac_comp": {"method": None},
                "temperature": {"method": None},
                "pressure": {"method": None},
                "flow_mol_phase": {"method": None},
                "dens_mol_phase": {"method": "_dens_mol_phase"},
                "pressure_sat": {"method": "_pressure_sat"},
                "mole_frac_phase_comp": {"method": "_mole_frac_phase"},
                "energy_internal_mol_phase_comp": {
                    "method": "_energy_internal_mol_phase_comp"
                },
                "energy_internal_mol_phase": {"method": "_energy_internal_mol_phase"},
                "enth_mol_phase_comp": {"method": "_enth_mol_phase_comp"},
                "enth_mol_phase": {"method": "_enth_mol_phase"},
                "entr_mol_phase_comp": {"method": "_entr_mol_phase_comp"},
                "entr_mol_phase": {"method": "_entr_mol_phase"},
                "temperature_bubble": {"method": "_temperature_bubble"},
                "temperature_dew": {"method": "_temperature_dew"},
                "pressure_bubble": {"method": "_pressure_bubble"},
                "pressure_dew": {"method": "_pressure_dew"},
                "fug_phase_comp": {"method": "_fug_phase_comp"},
            }
        )

        obj.define_custom_properties(
            {
                # Enthalpy of vaporization
                "dh_vap": {"method": "_dh_vap", "units": obj.derived_units.ENERGY_MOLE},
                # Entropy of vaporization
                "ds_vap": {
                    "method": "_ds_vap",
                    "units": obj.derived_units.ENTROPY_MOLE,
                },
            }
        )

        obj.add_default_units(
            {
                "time": pyunits.s,
                "length": pyunits.m,
                "mass": pyunits.kg,
                "amount": pyunits.mol,
                "temperature": pyunits.K,
            }
        )


class _IdealStateBlock(StateBlock):
    """
    This Class contains methods which should be applied to Property Blocks as a
    whole, rather than individual elements of indexed Property Blocks.
    """

    default_initializer = HDAInitializer

    def fix_initialization_states(blk):
        """
        Fixes state variables for state blocks.

        Returns:
            None
        """

        # Fix state variables
        fix_state_vars(blk)

        # Also need to deactivate sum of mole fraction constraint
        for k in blk.values():
            if not k.config.defined_state:
                k.equilibrium_constraint.deactivate()


@declare_process_block_class("IdealStateBlock", block_class=_IdealStateBlock)
class IdealStateBlockData(StateBlockData):
    """An example property package for ideal VLE."""

    def build(self):
        """Callable method for Block construction."""
        super().build()

        # Add state variables
        self.flow_mol_phase_comp = Var(
            self._params.phase_list,
            self._params.component_list,
            initialize=0.5,
            units=pyunits.mol / pyunits.s,
            bounds=(1e-12, 100),
            doc="Phase-component molar flow rates",
        )

        self.pressure = Var(
            initialize=101325,
            bounds=(100000, 1000000),
            units=pyunits.Pa,
            domain=NonNegativeReals,
            doc="State pressure",
        )
        self.temperature = Var(
            initialize=298.15,
            units=pyunits.K,
            bounds=(298, 1000),
            domain=NonNegativeReals,
            doc="State temperature",
        )

        # Add supporting variables
        def flow_mol_phase(b, p):
            return sum(b.flow_mol_phase_comp[p, j] for j in b._params.component_list)

        self.flow_mol_phase = Expression(
            self._params.phase_list, rule=flow_mol_phase, doc="Phase molar flow rates"
        )

        def flow_mol(b):
            return sum(
                b.flow_mol_phase_comp[p, j]
                for j in b._params.component_list
                for p in b._params.phase_list
            )

        self.flow_mol = Expression(rule=flow_mol, doc="Total molar flowrate")

        def mole_frac_phase_comp(b, p, j):
            return b.flow_mol_phase_comp[p, j] / b.flow_mol_phase[p]

        self.mole_frac_phase_comp = Expression(
            self._params.phase_list,
            self._params.component_list,
            rule=mole_frac_phase_comp,
            doc="Phase mole fractions",
        )

        def mole_frac_comp(b, j):
            return (
                sum(b.flow_mol_phase_comp[p, j] for p in b._params.phase_list)
                / b.flow_mol
            )

        self.mole_frac_comp = Expression(
            self._params.component_list,
            rule=mole_frac_comp,
            doc="Mixture mole fractions",
        )

        # Reaction Stoichiometry
        add_object_reference(
            self, "phase_equilibrium_list_ref", self._params.phase_equilibrium_list
        )

        if self.config.has_phase_equilibrium and self.config.defined_state is False:
            # Definition of equilibrium temperature for smooth VLE
            self._teq = Var(
                initialize=self.temperature.value,
                units=pyunits.K,
                doc="Temperature for calculating phase equilibrium",
            )
            self._t1 = Var(
                initialize=self.temperature.value,
                units=pyunits.K,
                doc="Intermediate temperature for calculating Teq",
            )

            self.eps_1 = Param(
                default=0.01,
                units=pyunits.K,
                mutable=True,
                doc="Smoothing parameter for Teq",
            )
            self.eps_2 = Param(
                default=0.0005,
                units=pyunits.K,
                mutable=True,
                doc="Smoothing parameter for Teq",
            )

            # PSE paper Eqn 13
            def rule_t1(b):
                return b._t1 == 0.5 * (
                    b.temperature
                    + b.temperature_bubble
                    + sqrt((b.temperature - b.temperature_bubble) ** 2 + b.eps_1**2)
                )

            self._t1_constraint = Constraint(rule=rule_t1)

            # PSE paper Eqn 14
            # TODO : Add option for supercritical extension
            def rule_teq(b):
                return b._teq == 0.5 * (
                    b._t1
                    + b.temperature_dew
                    - sqrt((b._t1 - b.temperature_dew) ** 2 + b.eps_2**2)
                )

            self._teq_constraint = Constraint(rule=rule_teq)

            def rule_tr_eq(b, i):
                return b._teq / b._params.temperature_crit[i]

            self._tr_eq = Expression(
                self._params.component_list,
                rule=rule_tr_eq,
                doc="Component reduced temperatures",
            )

            def rule_equilibrium(b, i):
                return b.fug_phase_comp["Liq", i] == b.fug_phase_comp["Vap", i]

            self.equilibrium_constraint = Constraint(
                self._params.component_list, rule=rule_equilibrium
            )

    # -----------------------------------------------------------------------------
    # Property Methods
    def _dens_mol_phase(self):
        self.dens_mol_phase = Var(
            self._params.phase_list,
            initialize=1.0,
            units=pyunits.mol * pyunits.m**-3,
            doc="Molar density",
        )

        def rule_dens_mol_phase(b, p):
            if p == "Vap":
                return b._dens_mol_vap()
            else:
                return b._dens_mol_liq()

        self.eq_dens_mol_phase = Constraint(
            self._params.phase_list, rule=rule_dens_mol_phase
        )

    def _energy_internal_mol_phase_comp(self):
        self.energy_internal_mol_phase_comp = Var(
            self._params.phase_list,
            self._params.component_list,
            units=pyunits.J / pyunits.mol,
            doc="Phase-component molar specific internal energies",
        )

        def rule_energy_internal_mol_phase_comp(b, p, j):
            if p == "Vap":
                return b.energy_internal_mol_phase_comp[p, j] == b.enth_mol_phase_comp[
                    p, j
                ] - const.gas_constant * (b.temperature - b._params.temperature_ref)
            else:
                return (
                    b.energy_internal_mol_phase_comp[p, j]
                    == b.enth_mol_phase_comp[p, j]
                )

        self.eq_energy_internal_mol_phase_comp = Constraint(
            self._params.phase_list,
            self._params.component_list,
            rule=rule_energy_internal_mol_phase_comp,
        )

    def _energy_internal_mol_phase(self):
        self.energy_internal_mol_phase = Var(
            self._params.phase_list,
            units=pyunits.J / pyunits.mol,
            doc="Phase molar specific internal energies",
        )

        def rule_energy_internal_mol_phase(b, p):
            return b.energy_internal_mol_phase[p] == sum(
                b.energy_internal_mol_phase_comp[p, i] * b.mole_frac_phase_comp[p, i]
                for i in b._params.component_list
            )

        self.eq_energy_internal_mol_phase = Constraint(
            self._params.phase_list, rule=rule_energy_internal_mol_phase
        )

    def _enth_mol_phase_comp(self):
        self.enth_mol_phase_comp = Var(
            self._params.phase_list,
            self._params.component_list,
            initialize=7e5,
            units=pyunits.J / pyunits.mol,
            doc="Phase-component molar specific enthalpies",
        )

        def rule_enth_mol_phase_comp(b, p, j):
            if p == "Vap":
                return b._enth_mol_comp_vap(j)
            else:
                return b._enth_mol_comp_liq(j)

        self.eq_enth_mol_phase_comp = Constraint(
            self._params.phase_list,
            self._params.component_list,
            rule=rule_enth_mol_phase_comp,
        )

    def _enth_mol_phase(self):
        self.enth_mol_phase = Var(
            self._params.phase_list,
            initialize=7e5,
            units=pyunits.J / pyunits.mol,
            doc="Phase molar specific enthalpies",
        )

        def rule_enth_mol_phase(b, p):
            return b.enth_mol_phase[p] == sum(
                b.enth_mol_phase_comp[p, i] * b.mole_frac_phase_comp[p, i]
                for i in b._params.component_list
            )

        self.eq_enth_mol_phase = Constraint(
            self._params.phase_list, rule=rule_enth_mol_phase
        )

    def _entr_mol_phase_comp(self):
        self.entr_mol_phase_comp = Var(
            self._params.phase_list,
            self._params.component_list,
            units=pyunits.J / pyunits.mol / pyunits.K,
            doc="Phase-component molar specific entropies",
        )

        def rule_entr_mol_phase_comp(b, p, j):
            if p == "Vap":
                return b._entr_mol_comp_vap(j)
            else:
                return b._entr_mol_comp_liq(j)

        self.eq_entr_mol_phase_comp = Constraint(
            self._params.phase_list,
            self._params.component_list,
            rule=rule_entr_mol_phase_comp,
        )

    def _entr_mol_phase(self):
        self.entr_mol_phase = Var(
            self._params.phase_list,
            units=pyunits.J / pyunits.mol / pyunits.K,
            doc="Phase molar specific enthropies",
        )

        def rule_entr_mol_phase(b, p):
            return b.entr_mol_phase[p] == sum(
                b.entr_mol_phase_comp[p, i] * b.mole_frac_phase_comp[p, i]
                for i in b._params.component_list
            )

        self.eq_entr_mol_phase = Constraint(
            self._params.phase_list, rule=rule_entr_mol_phase
        )

    # -----------------------------------------------------------------------------
    # General Methods
    def get_material_flow_terms(self, p, j):
        """Create material flow terms for control volume."""
        if not self.is_property_constructed("material_flow_terms"):
            try:

                def rule_material_flow_terms(blk, p, j):
                    return blk.flow_mol_phase_comp[p, j]

                self.material_flow_terms = Expression(
                    self.params.phase_list,
                    self.params.component_list,
                    rule=rule_material_flow_terms,
                )
            except AttributeError:
                self.del_component(self.material_flow_terms)

        if j in self.params.component_list:
            return self.material_flow_terms[p, j]
        else:
            return 0

    def get_enthalpy_flow_terms(self, p):
        """Create enthalpy flow terms."""
        if not self.is_property_constructed("enthalpy_flow_terms"):
            try:

                def rule_enthalpy_flow_terms(blk, p):
                    return blk.flow_mol_phase[p] * blk.enth_mol_phase[p]

                self.enthalpy_flow_terms = Expression(
                    self.params.phase_list, rule=rule_enthalpy_flow_terms
                )
            except AttributeError:
                self.del_component(self.enthalpy_flow_terms)
        return self.enthalpy_flow_terms[p]

    def get_material_density_terms(self, p, j):
        """Create material density terms."""
        if not self.is_property_constructed("material_density_terms"):
            try:

                def rule_material_density_terms(b, p, j):
                    return self.dens_mol_phase[p] * self.mole_frac_phase_comp[p, j]

                self.material_density_terms = Expression(
                    self.params.phase_list,
                    self.params.component_list,
                    rule=rule_material_density_terms,
                )
            except AttributeError:
                self.del_component(self.material_density_terms)

        if j in self.params.component_list:
            return self.material_density_terms[p, j]
        else:
            return 0

    def get_enthalpy_density_terms(self, p):
        """Create energy density terms."""
        if not self.is_property_constructed("enthalpy_density_terms"):
            try:

                def rule_energy_density_terms(b, p):
                    return self.dens_mol_phase[p] * self.energy_internal_mol_phase[p]

                self.energy_density_terms = Expression(
                    self.params.phase_list, rule=rule_energy_density_terms
                )
            except AttributeError:
                self.del_component(self.energy_density_terms)
        return self.enthalpy_density_terms[p]

    def default_material_balance_type(self):
        return MaterialBalanceType.componentPhase

    def default_energy_balance_type(self):
        return EnergyBalanceType.enthalpyTotal

    def get_material_flow_basis(b):
        return MaterialFlowBasis.molar

    def define_state_vars(self):
        """Define state vars."""
        return {
            "flow_mol_phase_comp": self.flow_mol_phase_comp,
            "temperature": self.temperature,
            "pressure": self.pressure,
        }

    # Property package utility functions
    def calculate_bubble_point_temperature(self, clear_components=True):
        """ "To compute the bubble point temperature of the mixture."""

        if hasattr(self, "eq_temperature_bubble"):
            # Do not delete components if the block already has the components
            clear_components = False

        calculate_variable_from_constraint(
            self.temperature_bubble, self.eq_temperature_bubble
        )

        return self.temperature_bubble.value

        if clear_components is True:
            self.del_component(self.eq_temperature_bubble)
            self.del_component(self._p_sat_bubbleT)
            self.del_component(self.temperature_bubble)

    def calculate_dew_point_temperature(self, clear_components=True):
        """ "To compute the dew point temperature of the mixture."""

        if hasattr(self, "eq_temperature_dew"):
            # Do not delete components if the block already has the components
            clear_components = False

        calculate_variable_from_constraint(
            self.temperature_dew, self.eq_temperature_dew
        )

        return self.temperature_dew.value

        # Delete the var/constraint created in this method that are part of the
        # IdealStateBlock if the user desires
        if clear_components is True:
            self.del_component(self.eq_temperature_dew)
            self.del_component(self._p_sat_dewT)
            self.del_component(self.temperature_dew)

    def calculate_bubble_point_pressure(self, clear_components=True):
        """ "To compute the bubble point pressure of the mixture."""

        if hasattr(self, "eq_pressure_bubble"):
            # Do not delete components if the block already has the components
            clear_components = False

        calculate_variable_from_constraint(
            self.pressure_bubble, self.eq_pressure_bubble
        )

        return self.pressure_bubble.value

        # Delete the var/constraint created in this method that are part of the
        # IdealStateBlock if the user desires
        if clear_components is True:
            self.del_component(self.eq_pressure_bubble)
            self.del_component(self._p_sat_bubbleP)
            self.del_component(self.pressure_bubble)

    def calculate_dew_point_pressure(self, clear_components=True):
        """ "To compute the dew point pressure of the mixture."""

        if hasattr(self, "eq_pressure_dew"):
            # Do not delete components if the block already has the components
            clear_components = False

        calculate_variable_from_constraint(self.pressure_dew, self.eq_pressure_dew)

        return self.pressure_dew.value

        # Delete the var/constraint created in this method that are part of the
        # IdealStateBlock if the user desires
        if clear_components is True:
            self.del_component(self.eq_pressure_dew)
            self.del_component(self._p_sat_dewP)
            self.del_component(self.pressure_dew)

    # -----------------------------------------------------------------------------
    # Bubble and Dew Points
    # Ideal-Ideal properties allow for the simplifications below
    # Other methods require more complex equations with shadow compositions

    # For future work, propose the following:
    # Core class writes a set of constraints Phi_L_i == Phi_V_i
    # Phi_L_i and Phi_V_i make calls to submethods which add shadow compositions
    # as needed
    def _temperature_bubble(self):
        self.temperature_bubble = Param(
            initialize=33.0, units=pyunits.K, doc="Bubble point temperature"
        )

    def _temperature_dew(self):

        self.temperature_dew = Var(
            initialize=298.15, units=pyunits.K, doc="Dew point temperature"
        )

        def rule_psat_dew(b, j):
            return (
                1e5
                * pyunits.Pa
                * 10
                ** (
                    b._params.pressure_sat_coeff_A[j]
                    - b._params.pressure_sat_coeff_B[j]
                    / (b.temperature_dew + b._params.pressure_sat_coeff_C[j])
                )
            )

        try:
            # Try to build expression
            self._p_sat_dewT = Expression(
                self._params.component_list, rule=rule_psat_dew
            )

            def rule_temp_dew(b):
                return (
                    b.pressure
                    * sum(
                        b.mole_frac_comp[i] / b._p_sat_dewT[i]
                        for i in ["toluene", "benzene"]
                    )
                    - 1
                    == 0
                )

            self.eq_temperature_dew = Constraint(rule=rule_temp_dew)
        except AttributeError:
            # If expression fails, clean up so that DAE can try again later
            # Deleting only var/expression as expression construction will fail
            # first; if it passes then constraint construction will not fail.
            self.del_component(self.temperature_dew)
            self.del_component(self._p_sat_dewT)

    def _pressure_bubble(self):
        self.pressure_bubble = Param(
            initialize=1e8, units=pyunits.Pa, doc="Bubble point pressure"
        )

    def _pressure_dew(self):
        self.pressure_dew = Var(
            initialize=298.15, units=pyunits.Pa, doc="Dew point pressure"
        )

        def rule_psat_dew(b, j):
            return (
                1e5
                * pyunits.Pa
                * 10
                ** (
                    b._params.pressure_sat_coeff_A[j]
                    - b._params.pressure_sat_coeff_B[j]
                    / (b.temperature + b._params.pressure_sat_coeff_C[j])
                )
            )

        try:
            # Try to build expression
            self._p_sat_dewP = Expression(
                self._params.component_list, rule=rule_psat_dew
            )

            def rule_pressure_dew(b):
                return (
                    b.pressure_dew
                    * sum(
                        b.mole_frac_comp[i] / b._p_sat_dewP[i]
                        for i in ["toluene", "benzene"]
                    )
                    - 1
                    == 0
                )

            self.eq_pressure_dew = Constraint(rule=rule_pressure_dew)
        except AttributeError:
            # If expression fails, clean up so that DAE can try again later
            # Deleting only var/expression as expression construction will fail
            # first; if it passes then constraint construction will not fail.
            self.del_component(self.pressure_dew)
            self.del_component(self._p_sat_dewP)

    # -----------------------------------------------------------------------------
    # Liquid phase properties
    def _dens_mol_liq(b):
        return b.dens_mol_phase["Liq"] == 1e3 * sum(
            b.mole_frac_phase_comp["Liq", j]
            * b._params.dens_liq_param_1[j]
            / b._params.dens_liq_param_2[j]
            ** (
                1
                + (1 - b.temperature / b._params.dens_liq_param_3[j])
                ** b._params.dens_liq_param_4[j]
            )
            for j in ["benzene", "toluene"]
        )

    def _fug_phase_comp(self):
        def fug_phase_comp_rule(b, p, i):
            if p == "Liq":
                if i in ["hydrogen", "methane"]:
                    return b.mole_frac_phase_comp["Liq", i]
                else:
                    return b.pressure_sat[i] * b.mole_frac_phase_comp["Liq", i]
            else:
                if i in ["hydrogen", "methane"]:
                    return 1e-6
                else:
                    return b.mole_frac_phase_comp["Vap", i] * b.pressure

        self.fug_phase_comp = Expression(
            self._params.phase_list,
            self._params.component_list,
            rule=fug_phase_comp_rule,
        )

    def _pressure_sat(self):
        self.pressure_sat = Var(
            self._params.component_list,
            initialize=101325,
            units=pyunits.Pa,
            doc="Vapor pressure",
        )

        def rule_P_sat(b, j):
            return (
                (
                    log10(b.pressure_sat[j] / pyunits.Pa * 1e-5)
                    - b._params.pressure_sat_coeff_A[j]
                )
                * (b._teq + b._params.pressure_sat_coeff_C[j])
            ) == -b._params.pressure_sat_coeff_B[j]

        self.eq_pressure_sat = Constraint(self._params.component_list, rule=rule_P_sat)

    def _enth_mol_comp_liq(b, j):
        return b.enth_mol_phase_comp["Liq", j] * 1e3 == (
            (b._params.cp_ig_5["Liq", j] / 5)
            * (b.temperature**5 - b._params.temperature_ref**5)
            + (b._params.cp_ig_4["Liq", j] / 4)
            * (b.temperature**4 - b._params.temperature_ref**4)
            + (b._params.cp_ig_3["Liq", j] / 3)
            * (b.temperature**3 - b._params.temperature_ref**3)
            + (b._params.cp_ig_2["Liq", j] / 2)
            * (b.temperature**2 - b._params.temperature_ref**2)
            + b._params.cp_ig_1["Liq", j] * (b.temperature - b._params.temperature_ref)
        )

    def _entr_mol_comp_liq(b, j):
        return b.entr_mol_phase_comp["Liq", j] * 1e3 == (
            (
                (b._params.cp_ig_5["Liq", j] / 4)
                * (b.temperature**4 - b._params.temperature_ref**4)
                + (b._params.cp_ig_4["Liq", j] / 3)
                * (b.temperature**3 - b._params.temperature_ref**3)
                + (b._params.cp_ig_3["Liq", j] / 2)
                * (b.temperature**2 - b._params.temperature_ref**2)
                + b._params.cp_ig_2["Liq", j]
                * (b.temperature - b._params.temperature_ref)
                + b._params.cp_ig_1["Liq", j]
                * log(b.temperature / b._params.temperature_ref)
            )
            - const.gas_constant
            * log(
                b.mole_frac_phase_comp["Liq", j] * b.pressure / b._params.pressure_ref
            )
        )

    # -----------------------------------------------------------------------------
    # Vapour phase properties
    def _dens_mol_vap(b):
        return b.pressure == (
            b.dens_mol_phase["Vap"] * const.gas_constant * b.temperature
        )

    def _dh_vap(self):
        # heat of vaporization
        add_object_reference(self, "dh_vap", self._params.dh_vap)

    def _ds_vap(self):
        # entropy of vaporization = dh_Vap/T_boil
        # TODO : something more rigorous would be nice
        self.ds_vap = Var(
            self._params.component_list,
            initialize=86,
            units=pyunits.J / pyunits.mol / pyunits.K,
            doc="Entropy of vaporization",
        )

        def rule_ds_vap(b, j):
            return b.dh_vap[j] == (b.ds_vap[j] * b._params.temperature_boil[j])

        self.eq_ds_vap = Constraint(self._params.component_list, rule=rule_ds_vap)

    def _enth_mol_comp_vap(b, j):
        return b.enth_mol_phase_comp["Vap", j] == b.dh_vap[j] + (
            (b._params.cp_ig_5["Vap", j] / 5)
            * (b.temperature**5 - b._params.temperature_ref**5)
            + (b._params.cp_ig_4["Vap", j] / 4)
            * (b.temperature**4 - b._params.temperature_ref**4)
            + (b._params.cp_ig_3["Vap", j] / 3)
            * (b.temperature**3 - b._params.temperature_ref**3)
            + (b._params.cp_ig_2["Vap", j] / 2)
            * (b.temperature**2 - b._params.temperature_ref**2)
            + b._params.cp_ig_1["Vap", j] * (b.temperature - b._params.temperature_ref)
        )

    def _entr_mol_comp_vap(b, j):
        return b.entr_mol_phase_comp["Vap", j] == (
            b.ds_vap[j]
            + (
                (b._params.cp_ig_5["Vap", j] / 4)
                * (b.temperature**4 - b._params.temperature_ref**4)
                + (b._params.cp_ig_4["Vap", j] / 3)
                * (b.temperature**3 - b._params.temperature_ref**3)
                + (b._params.cp_ig_3["Vap", j] / 2)
                * (b.temperature**2 - b._params.temperature_ref**2)
                + b._params.cp_ig_2["Vap", j]
                * (b.temperature - b._params.temperature_ref)
                + b._params.cp_ig_1["Vap", j]
                * log(b.temperature / b._params.temperature_ref)
            )
            - const.gas_constant
            * log(
                b.mole_frac_phase_comp["Vap", j] * b.pressure / b._params.pressure_ref
            )
        )

    def calculate_scaling_factors(self):
        # Get default scale factors
        super().calculate_scaling_factors()

        is_two_phase = len(self._params.phase_list) == 2
        sf_flow = iscale.get_scaling_factor(self.flow_mol, default=1, warning=True)
        sf_T = iscale.get_scaling_factor(self.temperature, default=1, warning=True)
        sf_P = iscale.get_scaling_factor(self.pressure, default=1, warning=True)

        if self.is_property_constructed("_teq"):
            iscale.set_scaling_factor(self._teq, sf_T)
        if self.is_property_constructed("_teq_constraint"):
            iscale.constraint_scaling_transform(
                self._teq_constraint, sf_T, overwrite=False
            )

        if self.is_property_constructed("_t1"):
            iscale.set_scaling_factor(self._t1, sf_T)
        if self.is_property_constructed("_t1_constraint"):
            iscale.constraint_scaling_transform(
                self._t1_constraint, sf_T, overwrite=False
            )

        if self.is_property_constructed("_mole_frac_pdew"):
            iscale.set_scaling_factor(self._mole_frac_pdew, 1e3)
            iscale.constraint_scaling_transform(
                self._sum_mole_frac_pdew, 1e3, overwrite=False
            )

        if self.is_property_constructed("total_flow_balance"):
            iscale.constraint_scaling_transform(
                self.total_flow_balance, sf_flow, overwrite=False
            )

        if self.is_property_constructed("component_flow_balances"):
            for i, c in self.component_flow_balances.items():
                if is_two_phase:
                    s = iscale.get_scaling_factor(
                        self.mole_frac_comp[i], default=1, warning=True
                    )
                    s *= sf_flow
                    iscale.constraint_scaling_transform(c, s, overwrite=False)
                else:
                    s = iscale.get_scaling_factor(
                        self.mole_frac_comp[i], default=1, warning=True
                    )
                    iscale.constraint_scaling_transform(c, s, overwrite=False)

        if self.is_property_constructed("dens_mol_phase"):
            for c in self.eq_dens_mol_phase.values():
                iscale.constraint_scaling_transform(c, sf_P, overwrite=False)

        if self.is_property_constructed("dens_mass_phase"):
            for p, c in self.eq_dens_mass_phase.items():
                sf = iscale.get_scaling_factor(
                    self.dens_mass_phase[p], default=1, warning=True
                )
                iscale.constraint_scaling_transform(c, sf, overwrite=False)

        if self.is_property_constructed("enth_mol_phase"):
            for p, c in self.eq_enth_mol_phase.items():
                sf = iscale.get_scaling_factor(
                    self.enth_mol_phase[p], default=1, warning=True
                )
                iscale.constraint_scaling_transform(c, sf, overwrite=False)

        if self.is_property_constructed("enth_mol"):
            sf = iscale.get_scaling_factor(self.enth_mol, default=1, warning=True)
            sf *= sf_flow
            iscale.constraint_scaling_transform(self.eq_enth_mol, sf, overwrite=False)

        if self.is_property_constructed("entr_mol_phase"):
            for p, c in self.eq_entr_mol_phase.items():
                sf = iscale.get_scaling_factor(
                    self.entr_mol_phase[p], default=1, warning=True
                )
                iscale.constraint_scaling_transform(c, sf, overwrite=False)

        if self.is_property_constructed("entr_mol"):
            sf = iscale.get_scaling_factor(self.entr_mol, default=1, warning=True)
            sf *= sf_flow
            iscale.constraint_scaling_transform(self.eq_entr_mol, sf, overwrite=False)

        if self.is_property_constructed("gibbs_mol_phase"):
            for p, c in self.eq_gibbs_mol_phase.items():
                sf = iscale.get_scaling_factor(
                    self.gibbs_mol_phase[p], default=1, warning=True
                )
                iscale.constraint_scaling_transform(c, sf, overwrite=False)



import pytest

from ahuora_builder.custom.thermal_utility_systems.steam_user import SteamUser
import pyomo.environ as pyo
from pyomo.environ import (
    units as UNIT,
    value,
)
from idaes.core import FlowsheetBlock
from idaes.core.util.model_statistics import degrees_of_freedom
from property_packages.build_package import build_package

@pytest.fixture
def steam_user_base_case():
    def _make_case(
            steam_flow,
            steam_pressure,
            steam_temperature,
            cond_return_rate,
            cond_return_temperature,
            deltaT_subcool,
            heat_loss,
            pressure_loss,
            has_desuperheating=False, 
            bfw_temperature=0,
            deltaT_superheat=0,
        ):
        # This defines the base case for all tests
        m = pyo.ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)
        m.fs.water = build_package("helmholtz", ["water"], ["Liq", "Vap"])
       
       # Create Steam User
        m.fs.user = SteamUser(
            property_package=m.fs.water,
            has_desuperheating=has_desuperheating,
        )
        m.fs.user.inlet_steam.flow_mol.fix(
            steam_flow
        )
        m.fs.user.inlet_steam.pressure.fix(
            steam_pressure
        ) 
        m.fs.user.inlet_steam.enth_mol.fix(
            m.fs.water.htpx(
                T=steam_temperature, 
                p=m.fs.user.inlet_steam_state[0].pressure,
            )
        ) 
        m.fs.user.cond_return_rate.fix(
            cond_return_rate
        )
        m.fs.user.cond_return_temperature.fix(
            cond_return_temperature
        )        
        m.fs.user.deltaT_subcool.fix(
            deltaT_subcool
        )
        m.fs.user.heat_loss.fix(
            heat_loss
        )
        m.fs.user.pressure_loss.fix(
            pressure_loss
        )
        if has_desuperheating:
            m.fs.user.bfw_temperature.fix(
                bfw_temperature
            )
            m.fs.user.deltaT_superheat.fix(
                deltaT_superheat
            )

        dof = degrees_of_freedom(m.fs)
        if not(dof == 0):
            raise ValueError("DoF is not zero.")

        return m
    return _make_case

def assert_user_solution(
        m,
        steam_flow,
        steam_pressure,
        steam_temperature,
        cond_return_rate,
        cond_return_temperature,
        deltaT_subcool,
        heat_loss,
        pressure_loss,
        has_desuperheating=False, 
        bfw_temperature=0,
        deltaT_superheat=0,
        ):
    eps_rel = 1e-4 
    eps_abs = 1e-4

    # Check degrees of freedom
    assert degrees_of_freedom(m.fs) == 0
    
    # Initialize the model
    m.fs.user.initialize()
    
    # Solve the model
    opt = pyo.SolverFactory("ipopt")
    results = opt.solve(m, tee=False)
    
    # Check that solve was successful
    assert results.solver.termination_condition == pyo.TerminationCondition.optimal
    
    # Verify mass balance: total inlet flow = total vapor outlet flow + liquid outlet flow
    if value(steam_flow):
        assert 0 == pytest.approx(value(m.fs.user.inlet_steam_state[0].flow_mol - steam_flow), rel=eps_rel, abs=eps_abs)
    assert 0 == pytest.approx(value(m.fs.user.inlet_steam_state[0].pressure - steam_pressure), rel=eps_rel, abs=eps_abs)
    assert 0 == pytest.approx(value(m.fs.user.inlet_steam_state[0].temperature - steam_temperature), rel=eps_rel, abs=0.1)
    
    total_flow_in = sum([value(o[0].flow_mol) for o in m.fs.user.inlet_blocks])
    total_flow_out = sum([value(o[0].flow_mol) for o in m.fs.user.outlet_blocks])
    assert total_flow_in == pytest.approx(total_flow_out, rel=eps_rel, abs=eps_abs)

    assert value(cond_return_rate * total_flow_in) == pytest.approx(value(m.fs.user.outlet_return.flow_mol[0]), rel=eps_rel, abs=eps_abs)
    assert value(cond_return_temperature) == pytest.approx(value(m.fs.user.outlet_return_state[0].temperature), rel=eps_rel, abs=eps_abs)
    assert value(deltaT_subcool) == pytest.approx(value(m.fs.user._int_outlet_sat_liq_state[0].temperature - m.fs.user._int_outlet_cond_state[0].temperature), rel=eps_rel, abs=eps_abs)
    
    assert value(heat_loss) == pytest.approx(value(
            UNIT.convert(
                m.fs.user._int_mixed_inlet_state[0].flow_mol * m.fs.user._int_mixed_inlet_state[0].enth_mol - m.fs.user.heat_demand[0] - m.fs.user._int_outlet_cond_state[0].flow_mol * m.fs.user._int_outlet_cond_state[0].enth_mol, 
                to_units=UNIT.W
            )
        ), rel=eps_rel, abs=eps_abs
    )
    assert value(pressure_loss) == pytest.approx(value(
            UNIT.convert(
                m.fs.user._int_mixed_inlet_state[0].pressure - m.fs.user._int_outlet_cond_state[0].pressure,
                to_units=UNIT.Pa
            )
        ), 
        rel=eps_rel, abs=eps_abs
    )

    if has_desuperheating:
        assert value(bfw_temperature) == pytest.approx(value(m.fs.user.inlet_water_state[0].temperature), rel=eps_rel, abs=eps_abs)
        assert value(deltaT_superheat) == pytest.approx(value(m.fs.user._int_mixed_inlet_state[0].temperature - m.fs.user._int_inlet_sat_vap_state[0].temperature), rel=eps_rel, abs=eps_abs)

def test_steam_user_flow_case(steam_user_base_case):
    args = {
        "steam_flow": 1000 * UNIT.mol / UNIT.s,
        "steam_pressure": 2000 * 1000 * UNIT.Pa,
        "steam_temperature": (250 + 273.15) * UNIT.K,
        "cond_return_rate": 0.7,
        "cond_return_temperature": (80 + 273.15) * UNIT.K,
        "deltaT_subcool": 10 * UNIT.K,
        "heat_loss": 1000 * UNIT.W,
        "pressure_loss": 5000 * UNIT.Pa,
        "has_desuperheating": False,
        "bfw_temperature": (80 + 273.15) * UNIT.K,
        "deltaT_superheat": 10 * UNIT.K,        
    }
    m = steam_user_base_case(**args)
    assert_user_solution(m, **args)

def test_steam_user_flow_case_with_parital_desuperheating(steam_user_base_case):
    args = {
        "steam_flow": 1000 * UNIT.mol / UNIT.s,
        "steam_pressure": 2000 * 1000 * UNIT.Pa,
        "steam_temperature": (250 + 273.15) * UNIT.K,
        "cond_return_rate": 0.7,
        "cond_return_temperature": (80 + 273.15) * UNIT.K,
        "deltaT_subcool": 10 * UNIT.K,
        "heat_loss": 1000 * UNIT.W,
        "pressure_loss": 5000 * UNIT.Pa,
        "has_desuperheating": True,
        "bfw_temperature": (80 + 273.15) * UNIT.K,
        "deltaT_superheat": 10 * UNIT.K,        
    }
    m = steam_user_base_case(**args)
    assert_user_solution(m, **args)

def test_steam_user_flow_case_with_full_desuperheating(steam_user_base_case):
    args = {
        "steam_flow": 1000 * UNIT.mol / UNIT.s,
        "steam_pressure": 2000 * 1000 * UNIT.Pa,
        "steam_temperature": (250 + 273.15) * UNIT.K,
        "cond_return_rate": 0.7,
        "cond_return_temperature": (80 + 273.15) * UNIT.K,
        "deltaT_subcool": 10 * UNIT.K,
        "heat_loss": 1000 * UNIT.W,
        "pressure_loss": 5000 * UNIT.Pa,
        "has_desuperheating": True,
        "bfw_temperature": (80 + 273.15) * UNIT.K,
        "deltaT_superheat": 0 * UNIT.K,        
    }
    m = steam_user_base_case(**args)
    assert_user_solution(m, **args)

def test_steam_user_demand_case(steam_user_base_case):
    args = {
        "steam_flow": 1000 * UNIT.mol / UNIT.s,
        "steam_pressure": 2000 * 1000 * UNIT.Pa,
        "steam_temperature": (250 + 273.15) * UNIT.K,
        "cond_return_rate": 0.7,
        "cond_return_temperature": (80 + 273.15) * UNIT.K,
        "deltaT_subcool": 0 * UNIT.K,
        "heat_loss": 0 * UNIT.W,
        "pressure_loss": 0 * UNIT.Pa,
        "has_desuperheating": False,
        "bfw_temperature": (80 + 273.15) * UNIT.K,
        "deltaT_superheat": 10 * UNIT.K,        
    }
    m = steam_user_base_case(**args)
    m.fs.user.heat_demand.fix(
        5 * 1000 * 1000 * UNIT.W
    )
    m.fs.user.inlet_steam.flow_mol[0].unfix()
    args["steam_flow"] = False
    assert_user_solution(m, **args)

def test_steam_user_demand_case_with_subcooling(steam_user_base_case):
    args = {
        "steam_flow": 0 * UNIT.mol / UNIT.s,
        "steam_pressure": 2000 * 1000 * UNIT.Pa,
        "steam_temperature": (250 + 273.15) * UNIT.K,
        "cond_return_rate": 0.7,
        "cond_return_temperature": (80 + 273.15) * UNIT.K,
        "deltaT_subcool": 10 * UNIT.K,
        "heat_loss": 0 * UNIT.W,
        "pressure_loss": 0 * UNIT.Pa,
        "has_desuperheating": False,
        "bfw_temperature": (80 + 273.15) * UNIT.K,
        "deltaT_superheat": 10 * UNIT.K,        
    }
    m = steam_user_base_case(**args)
    m.fs.user.heat_demand.fix(
        5 * 1000 * 1000 * UNIT.W
    )
    m.fs.user.inlet_steam.flow_mol[0].unfix()
    args["steam_flow"] = False
    assert_user_solution(m, **args)

def test_steam_user_demand_case_with_subcooling_and_desuperheating(steam_user_base_case):
    args = {
        "steam_flow": 0 * UNIT.mol / UNIT.s,
        "steam_pressure": 2000 * 1000 * UNIT.Pa,
        "steam_temperature": (250 + 273.15) * UNIT.K,
        "cond_return_rate": 0.5,
        "cond_return_temperature": (80 + 273.15) * UNIT.K,
        "deltaT_subcool": 10 * UNIT.K,
        "heat_loss": 1000 * UNIT.W,
        "pressure_loss": 1000 * UNIT.Pa,
        "has_desuperheating": True,
        "bfw_temperature": (110 + 273.15) * UNIT.K,
        "deltaT_superheat": 10 * UNIT.K,        
    }
    m = steam_user_base_case(**args)
    m.fs.user.heat_demand.fix(
        5 * 1000 * 1000 * UNIT.W
    )
    m.fs.user.inlet_steam.flow_mol[0].unfix()
    args["steam_flow"] = False
    assert_user_solution(m, **args)

def test_steam_user_demand_case_with_zero_flow(steam_user_base_case):
    args = {
        "steam_flow": 0 * UNIT.mol / UNIT.s,
        "steam_pressure": 2000 * 1000 * UNIT.Pa,
        "steam_temperature": (250 + 273.15) * UNIT.K,
        "cond_return_rate": 0.5,
        "cond_return_temperature": (80 + 273.15) * UNIT.K,
        "deltaT_subcool": 0 * UNIT.K,
        "heat_loss": 0 * UNIT.W,
        "pressure_loss": 0 * UNIT.Pa,
        "has_desuperheating": False,
        "bfw_temperature": (110 + 273.15) * UNIT.K,
        "deltaT_superheat": 10 * UNIT.K,        
    }
    m = steam_user_base_case(**args)
    args["steam_flow"] = False
    assert_user_solution(m, **args)



from idaes.models.unit_models import Compressor, Heater
from idaes.models.unit_models.heat_exchanger import delta_temperature_underwood_callback

from ahuora_builder_types import FlowsheetSchema
from ...generate_python_file import generate_python_code, PythonFileGenerator

GENERATE_TESTS = False

def test_generate_is_false():
    assert GENERATE_TESTS is False, "test_generate_python_file.py:  GENERATE_TESTS should be False to avoid overriding test_solved.json files during tests."


def test_import():
    generator = PythonFileGenerator({})
    heater_name, heater_import = generator.resolve_import(Heater)
    assert heater_name == "Heater"
    assert heater_import == "from idaes.models.unit_models.heater import Heater"
    compressor_name, compressor_import = generator.resolve_import(Compressor)
    assert compressor_name == "Compressor"
    assert compressor_import == "from idaes.models.unit_models.pressure_changer import Compressor"
    delta_t_name, delta_t_import = generator.resolve_import(delta_temperature_underwood_callback)
    assert delta_t_name == "delta_temperature_underwood_callback"
    assert delta_t_import == "from idaes.models.unit_models.heat_exchanger import delta_temperature_underwood_callback"


def test_names():
    # names need to be converted to valid python variable names
    generator = PythonFileGenerator({})
    def check_name(name: str, expected: str):
        # starts with m.fs.
        assert generator.get_name(name) == f"m.fs.{expected}"
    check_name("Pump-1", "Pump_1")
    check_name("Heater-1", "Heater_1")
    check_name("heat_exchanger-1", "heat_exchanger_1")
    check_name("name with spaces", "name_with_spaces")
    check_name("name-with-dashes", "name_with_dashes")
    check_name("name_with_underscores", "name_with_underscores")
    check_name("name_with_123_numbers", "name_with_123_numbers")
    check_name("123_numbers_at_start", "_123_numbers_at_start")
    check_name("name_with_!@#$%^&*()_special_characters", "name_with__special_characters")
    check_name("", "_unnamed_unit")
    check_name(" ", "_unnamed_unit")
    check_name("1", "_1")
    check_name(",", "_unnamed_unit")
    

def run_generate(input_file: str, expected_file: str) -> None:
    import os
    import json

    with open(os.path.join(os.path.dirname(__file__), input_file), "r") as f:
        flowsheet = json.load(f)
        flowsheet = FlowsheetSchema.model_validate(flowsheet)
    
    result = generate_python_code(flowsheet)

    # # write result to a file for debugging - or to regenerate expected files
    if GENERATE_TESTS:
        with open(os.path.join(os.path.dirname(__file__), expected_file), "w") as f:
            f.write(result)


    # split result line by line to compare each line
    lines = result.split("\n")

    with open(os.path.join(os.path.dirname(__file__), expected_file), "r") as f:
        expected_lines = f.readlines()
    
    # compare each line of the result with the expected file
    i = 0
    while len(lines) > 0 and len(expected_lines) > 0:
        # skip empty lines
        while len(lines) > 0 and lines[0].strip() == "":
            lines.pop(0)
        while len(expected_lines) > 0 and expected_lines[0].strip() == "":
            i += 1
            expected_lines.pop(0)
        if len(lines) == 0:
            assert len(expected_lines) == 0, f"Expected '{expected_lines[0]}'"
            break
        if len(expected_lines) == 0:
            assert len(lines) == 0, f"Got '{lines[0]}'"
            break
        # compare lines
        i += 1
        line = lines.pop(0).strip()
        expected_line = expected_lines.pop(0).strip()
        assert line == expected_line, f"Line {str(i)}: Expected '{expected_line}', got '{line}'"


def test_compressor():
    # basic test
    run_generate("../test_solver/configurations/compressor.json", "configurations/compressor_generated.py")


def test_heat_exchanger():
    # tests a few more features (constants, dictionaries)
    run_generate("../test_solver/configurations/heat_exchanger.json", "configurations/heat_exchanger_generated.py")


def test_pump_and_heater():
    # tests connected components
    run_generate("../test_solver/configurations/pump.json", "configurations/pump_generated.py")


def test_recycle():
    # tests tears
    run_generate("../test_solver/configurations/recycle.json", "configurations/recycle_generated.py")


from ..properties_manager import PropertiesManager
from typing import Iterator, TextIO
import sys
import pyomo.environ as pyo
from pyomo.core.base.constraint import ScalarConstraint

def compute_infeasibilities(properties_map: PropertiesManager) -> Iterator[tuple[any,float]]:
    """
    Compute how far each constrained property is from its target value.
    
    :param properties_map: PropertiesManager instance containing property components
    :type properties_map: PropertiesManager
    :return: A tuple list of property IDs and their infeasibility magnitudes
    :rtype: list[tuple[Any, float]]
    """

    for id, property_component in properties_map.items():
        if property_component.corresponding_constraint is not None and type(next(iter(property_component.corresponding_constraint))) is ScalarConstraint:
            # Assuming the component is a Pyomo Var or Expression
            current_value :float = pyo.value(next(iter(property_component.corresponding_constraint)))
            target_value :float = pyo.value(next(iter(property_component.corresponding_constraint)).upper)
            infeasibility = abs(current_value - target_value)
            yield (id, infeasibility)

def print_infeasibilities(properties_map: PropertiesManager):
    """
    Print the infeasibilities of constrained properties, sorted by magnitude.
    
    :param properties_map: PropertiesManager instance containing property components
    :type properties_map: PropertiesManager
    """


    # Sort from most to least infeasible
    infeasibilities = list(compute_infeasibilities(properties_map))
    infeasibilities.sort(key=lambda x: x[1], reverse=True)

    print("Property Infeasibilities:")
    for id, infeasibility in infeasibilities:
        print(f"Property: {properties_map.get(id).name}, Infeasibility: {infeasibility}")


from pyomo.network import Port
from ahuora_builder_types import PortId


class PortManager:
    
    def __init__(self) -> None:
        self._map: dict[PortId, Port] = {}
    

    def register_port(self, id: PortId, port: Port) -> None:
        """Registers a port with the port map, so that arcs can connect to it by id"""
        self._map[id] =  port
    
    
    def get_port(self, id: PortId) -> Port:
        try:
            return self._map[id]
        except KeyError:
            raise KeyError(f"Port with id `{id}` not found")


# Pyomo core
from pyomo.environ import (
    Constraint,
    Expression,
    NonNegativeReals,
    Suffix,
    Var,
    value,
    units as UNIT,
)
from pyomo.core.base.reference import Reference
from pyomo.common.config import ConfigBlock, ConfigValue, Bool

# IDAES core
from idaes.core import (
    declare_process_block_class,
    UnitModelBlockData,
    useDefault,
    StateBlock,
)
from idaes.core.util import scaling
from idaes.core.util.config import is_physical_parameter_block
from idaes.core.util.tables import create_stream_table_dataframe
from idaes.core.solvers import get_solver
from idaes.core.initialization import ModularInitializerBase

# Other
from property_packages.build_package import build_package

# Logger
import idaes.logger as idaeslog

# Typing
from typing import List


__author__ = "Ahuora Centre for Smart Energy Systems, University of Waikato, New Zealand"

# Set up logger
_log = idaeslog.getLogger(__name__)

class SteamUserInitializer(ModularInitializerBase):
    """Initializer for ``SteamUser``.

    Parameters
    ----------
    blk : SteamUser
        The unit model block to initialize.
    solver : optional
        A Pyomo/IDAES solver instance. Defaults to :func:`idaes.core.solvers.get_solver`.
    solver_options : dict, optional
        Options to set on the solver, e.g. tolerances.
    outlvl : int, optional
        IDAES log level (e.g. :data:`idaes.logger.WARNING`).

    Returns
    -------
    pyomo.opt.results.results_.SolverResults
        Result from the final solve.
    """

    def initialize(self, blk, **kwargs):
        # --- Solver setup
        solver = kwargs.get("solver", None) or get_solver()
        solver_options = kwargs.get("solver_options", {})
        for k, v in solver_options.items():
            solver.options[k] = v

        outlvl = kwargs.get("outlvl", idaeslog.WARNING)
        log = idaeslog.getLogger(__name__)

        # --- Time index
        t0 = blk.flowsheet().time.first()

        # --- 1) Initialize inlet state blocks
        inlet_blocks = list(blk.inlet_blocks)
        if blk.config.has_desuperheating:
            inlet_steam, inlet_water = inlet_blocks
            inlet_water[t0].pressure.set_value(
                inlet_steam[t0].pressure
            )
            inlet_water[t0].enth_mol.set_value(
                blk.water.htpx(
                    blk.bfw_temperature[t0], 
                    inlet_water[t0].pressure,
                ) 
            )
        else:
            inlet_steam = inlet_blocks[0]

        for sb in inlet_blocks:
            if hasattr(sb, "initialize"):
                sb.initialize(outlvl=outlvl)

        # --- 2) Seed satuarate inlet-related state
        if blk.config.has_desuperheating:
            sb = blk._int_inlet_sat_vap_state
            sb[t0].pressure.set_value(
                inlet_steam[t0].pressure
            )
            sb[t0].enth_mol.set_value(
                inlet_steam[t0].enth_mol_sat_phase["Vap"]
            )
            if hasattr(sb, "initialize"):
                sb.initialize(outlvl=outlvl)

        # --- 3) Aggregate inlet info for seeding mixed state block
        ms = blk._int_mixed_inlet_state
        ms[t0].pressure.set_value(
            inlet_steam[t0].pressure
        )
        if blk.config.has_desuperheating:
            if value(blk.deltaT_superheat[t0]) > 0:
                ms[t0].enth_mol.set_value(
                    blk.water.htpx(
                        T=blk._int_inlet_sat_vap_state[t0].temperature + blk.deltaT_superheat[t0],
                        p=inlet_steam[t0].pressure,
                    )            
                )
            else:
                ms[t0].enth_mol.set_value(
                    blk._int_inlet_sat_vap_state[t0].enth_mol         
                )     

            if abs(value(ms[t0].enth_mol - inlet_water[t0].enth_mol)) > 0:
                ms[t0].flow_mol.set_value(
                    inlet_steam[t0].flow_mol * (inlet_steam[t0].enth_mol - ms[t0].enth_mol) / (ms[t0].enth_mol - inlet_water[t0].enth_mol)
                )
            else:
                ms[t0].flow_mol.set_value(
                    inlet_steam[t0].flow_mol
                )
        else:
            ms[t0].enth_mol.set_value(
                inlet_steam[t0].enth_mol         
            )            
            ms[t0].flow_mol.set_value(
                inlet_steam[t0].flow_mol
            )

        if hasattr(ms, "initialize"):
            ms.initialize(outlvl=outlvl)

        # --- 4) Seed outlet-related states
        # Condensate after heating (internal)
        ios = blk._int_outlet_cond_state
        ios[t0].flow_mol.set_value(
            ms[t0].flow_mol
        )
        ios[t0].pressure.set_value(
            ms[t0].pressure - blk.pressure_loss[t0]
        )
        ios[t0].enth_mol.set_value(
            blk.water.htpx(
                T=blk._int_outlet_cond_state[t0].temperature_sat - blk.deltaT_subcool[t0],
                p=ios[t0].pressure,
            )
        )
        if hasattr(ios, "initialize"):
            ios.initialize(outlvl=outlvl)

        # --- 5) Seed satuarate outlet-related state
        sb = blk._int_outlet_sat_liq_state
        sb[t0].pressure.set_value(
            ios[t0].pressure
        )
        sb[t0].enth_mol.set_value(
            ios[t0].enth_mol_sat_phase["Vap"]
        )
        if hasattr(sb, "initialize"):
            sb.initialize(outlvl=outlvl)           

        # --- 5) Seed external outlets
        # Return line properties at user-specified temperature
        ors = blk.outlet_return_state
        ors[t0].flow_mol.set_value(ms[t0].flow_mol * value(blk.cond_return_rate[t0]))
        ors[t0].pressure.set_value(
            ms[t0].pressure - blk.pressure_loss[t0]
        )
        ors[t0].enth_mol.set_value(
            blk.water.htpx(
                T=blk.cond_return_temperature[t0], 
                p=ors[t0].pressure,
            )
        )
        if hasattr(ors, "initialize"):
            ors.initialize(outlvl=outlvl)

        # --- 6) Drain/blowdown outlet uses reference enthalpy (fixed later in build)
        # Flow/pressure will be solved by constraints
        if hasattr(blk.outlet_drain_state, "initialize"):
            blk.outlet_drain_state.initialize(outlvl=outlvl)

        # --- 7) Solve
        res = solver.solve(blk, tee=False)
        log.info(f"SteamUser init status: {res.solver.termination_condition}")
        return res

def _make_config_block(config):
    """Declare configuration options for the SteamUser unit.

    Declares property package references and integer counts for inlets and outlets.

    Args:
        config (ConfigBlock): The mutable configuration block to populate.
    """

    config.declare(
        "property_package",
        ConfigValue(
            default=useDefault,
            domain=is_physical_parameter_block,
            description="Property package to use for control volume",
        ),
    )
    config.declare(
        "property_package_args",
        ConfigBlock(
            implicit=True,
            description="Arguments to use for constructing property packages",
        ),
    )
    config.declare(
        "has_desuperheating",
        ConfigValue(
            default=False,
            domain=Bool,
            description="If true, include desuperheating prior to use as process heat. " \
            "Adds the state variable of the degree of superheat after desuperheating. Default: 0.",
        ),
    )    

@declare_process_block_class("SteamUser")
class SteamUserData(UnitModelBlockData):
    """Steam user unit operation.

    The SteamUser aggregates thermal loads from multiple sub-users (heaters) within a site.
    Desuperheating the flow is optional. Heat loss and pressure loss may be defined. 
    A mixed (intermediate) states are used for balances.

    Key features:
        - Material, energy, and momentum balances around the user
        - Optional desuperheating step prior to process use
        - User-specified condensate return rate and temperature
        - Optional heat and pressure losses.

    Attributes:
        inlet_list (list[str]): Names for inlet ports.
        outlet_list (list[str]): Names for outlet ports (incl. condensate/ and vent).
        inlet_blocks (list): StateBlocks for all inlets.
        outlet_blocks (list): StateBlocks for all outlets.
        _int_mixed_inlet_state: Intermediate mixture StateBlock.
        
        heat_loss (Var): Heat loss from the header (W).
        pressure_loss (Var): Pressure drop from inlet minimum to mixed state (Pa).
        bfw_flow_mol (Var): Required inlet boiler feed water flow for desuperheating (mol/s).
    """

    default_initializer=SteamUserInitializer
    CONFIG = UnitModelBlockData.CONFIG()
    _make_config_block(CONFIG)

    def build(self) -> None:
        """Build the unit model structure (ports, states, constraints)."""
        # 1. Inherit standard UnitModelBlockData properties and functions
        super().build()

        # 2. Validate input parameters are valid
        self._validate_model_config()

        # 3. Create lists of ports with state blocks to add
        self.inlet_list = self._create_inlet_port_name_list()
        self.outlet_list = self._create_outlet_port_name_list()

        # 4. Declare ports, state blocks and state property bounds 
        self.inlet_blocks = self._add_ports_with_state_blocks(
            stream_list=self.inlet_list,
            is_inlet=True,
            has_phase_equilibrium=False,
            is_defined_state=True,
        )
        self.outlet_blocks = self._add_ports_with_state_blocks(
            stream_list=self.outlet_list,
            is_inlet=False,
            has_phase_equilibrium=False,
            is_defined_state=False
        )
        self._internal_blocks = self._add_internal_state_blocks()
        self._ref_enth = self._add_environmental_reference_enth()
        self._add_bounds_to_state_properties()
        
        # 4. Declare references, variables and expressions for external and internal use
        self._create_references()
        self._create_variables()
        self._create_expressions()

        # 5. Set balance equations
        self._add_material_balances()
        self._add_energy_balances()
        self._add_momentum_balances()
        self._add_additional_constraints()

        # 6. Other
        self.scaling_factor = Suffix(direction=Suffix.EXPORT)

    # ------------------------------------------------------------------
    # Helpers & construction utilities
    # ------------------------------------------------------------------
    def _validate_model_config(self) -> bool:
        """Validate configuration for inlet and outlet counts.

        Raises:
            ValueError: If ``property_package is None``.
        """
        if self.config.property_package is None:
            raise ValueError("SteamUser: Property package not defined.")
        return True

    def _create_inlet_port_name_list(self) -> List[str]:
        """Build ordered inlet port names.

        Returns:
            list[str]: Names
        """
        
        return (
            [
                "inlet_steam", "inlet_water"
            ] 
            if self.config.has_desuperheating else
            [
                "inlet_steam"
            ]
        )            

    def _create_outlet_port_name_list(self) -> List[str]:
        """Build ordered outlet port names.

        Returns:
            list[str]: Names 
        """        
        return [
            "outlet_return", 
            "outlet_drain",
        ]

    def _add_ports_with_state_blocks(self, 
                                     stream_list: List[str], 
                                     is_inlet: List[str], 
                                     has_phase_equilibrium: bool=False,
                                     is_defined_state: bool=None,
                                     ) -> List[StateBlock]:
        """Construct StateBlocks and expose them as ports.

        Creates a StateBlock per named stream and attaches a corresponding inlet or
        outlet Port. Inlet blocks are defined states; outlet blocks are calculated states.

        Args:
            stream_list (list[str]): Port/StateBlock base names to create.
            is_inlet (bool): If True, create inlet ports with ``defined_state=True``;
                otherwise create outlet ports with ``defined_state=False``.
            has_phase_equilibrium (bool)

        Returns:
            list: The created StateBlocks, in the same order as ``stream_list``.
        """
        # Create empty list to hold StateBlocks for return
        state_block_ls = []

        # Setup StateBlock argument dict
        tmp_dict = dict(**self.config.property_package_args)
        tmp_dict["has_phase_equilibrium"] = has_phase_equilibrium
        if is_defined_state == None:
            tmp_dict["defined_state"] = True if is_inlet else False
        else:
            tmp_dict["defined_state"] = is_defined_state

        # Create an instance of StateBlock for all streams
        for s in stream_list:
            sb = self.config.property_package.build_state_block(
                self.flowsheet().time, doc=f"Thermophysical properties at {s}", **tmp_dict
            )
            setattr(
                self, s + "_state", 
                sb
            )
            state_block_ls.append(sb)
            add_fn = self.add_inlet_port if is_inlet else self.add_outlet_port
            add_fn(
                name=s,
                block=sb,
            )

        return state_block_ls

    def _add_internal_state_blocks(self) -> List[StateBlock]:
        """Create the intermediate StateBlock(s)."""
        # The _int_outlet_cond_state:
        #     - Is not a defined state (solved from balances).
        #     - Represents the state of the condensate after delivering process heating.
        tmp_dict = dict(**self.config.property_package_args)
        tmp_dict["has_phase_equilibrium"] = False
        tmp_dict["defined_state"] = False

        self._int_outlet_cond_state = self.config.property_package.build_state_block(
            self.flowsheet().time, 
            doc="Thermophysical properties of condensate after process heating.", 
            **tmp_dict
        )
        # The _int_mixed_inlet_state:
        #     - Has phase equilibrium enabled.
        #     - Is not a defined state (solved from balances).
        #     - Always exists even when not desuperheating
        tmp_dict["has_phase_equilibrium"] = True
        tmp_dict["defined_state"] = False            
        self._int_mixed_inlet_state = self.config.property_package.build_state_block(
            self.flowsheet().time, 
            doc="Thermophysical properties internal mixed inlet state after desuperheating (if applicable).", 
            **tmp_dict
        )
        # The _int_outlet_sat_liq_state:
        #     - Has phase equilibrium enabled.
        #     - Is not a defined state (solved from balances).
        #     - Always exists even when not desuperheating
        tmp_dict["has_phase_equilibrium"] = True
        tmp_dict["defined_state"] = False            
        self._int_outlet_sat_liq_state = self.config.property_package.build_state_block(
            self.flowsheet().time, 
            doc="Thermophysical properties internal mixed saturate state.", 
            **tmp_dict
        )
        self._int_outlet_sat_liq_state[:].flow_mol.fix(1)    

        if self.config.has_desuperheating:
            # The _int_inlet_sat_vap_state:
            #     - Has phase equilibrium enabled.
            #     - Is not a defined state (solved from balances).
            #     - Only exists when desuperheating
            tmp_dict["has_phase_equilibrium"] = True
            tmp_dict["defined_state"] = False            
            self._int_inlet_sat_vap_state = self.config.property_package.build_state_block(
                self.flowsheet().time, 
                doc="Thermophysical properties internal mixed saturate state.", 
                **tmp_dict
            )
            self._int_inlet_sat_vap_state[:].flow_mol.fix(1)

        return [
            self._int_mixed_inlet_state,
            self._int_outlet_cond_state,
            self._int_outlet_sat_liq_state,
            self._int_inlet_sat_vap_state,
        ] if self.config.has_desuperheating else [
            self._int_mixed_inlet_state,
            self._int_outlet_cond_state,
            self._int_outlet_sat_liq_state,
        ]      
    
    def _add_environmental_reference_enth(self) -> None:
        """Create a helper to compute reference enthalpy at 15°C, 1 atm (water)."""
        self.water = build_package("helmholtz", ["water"], ["Liq"])
        return self.water.htpx(
            (15 + 273.15) * UNIT.K, 
            101325 * UNIT.Pa
        )

    def _add_bounds_to_state_properties(self) -> None:
        """Add lower and/or upper bounds to state properties.

        - Set nonnegativity lower bounds on all inlet/intermediate/outlet flows.
        """
        for sb in (self.inlet_blocks + self.outlet_blocks + self._internal_blocks):
            for t in sb:
                sb[t].flow_mol.setlb(0.0)     

    def _create_references(self) -> None:
        """Create convenient References.

        Creates references to _int_mixed_inlet_state properties:
            - ``bfw_temperature`` 
            - ``bfw_flow_mass``
            - ``bfw_flow_mol``
        """
        # Read only variables, only applicable if desuperheating is active
        if self.config.has_desuperheating:
            self.bfw_flow_mass = Reference(
                self.inlet_water_state[:].flow_mass
            )
            self.bfw_flow_mol = Reference(
                self.inlet_water_state[:].flow_mol
            )
            self.inlet_water_state[:].flow_mol.unfix()  

    def _create_variables(self) -> None:
        """Declare decision/parameter variables for the unit.

        Creates:
            - ``heat_demand`` 
            - ``cond_return_rate`` 
            - ``cond_return_temperature`` 
            - ``deltaT_subcool`` 
            - ``heat_loss`` 
            - ``pressure_loss`` 
            If desuperheating:
                - ``bfw_temperature``  
                - ``deltaT_superheat``      
        """
        # Get units consistent with the property package
        units_meta = self.config.property_package.get_metadata()

        # Calculated: Process heat demand (kW) — user typically fixes
        self.heat_demand = Var(
            self.flowsheet().time, 
            domain=NonNegativeReals,
            doc="Process heat demand of the users. Default: 0 kW.",
            units=units_meta.get_derived_units("power"),
        )
        self.heat_demand[:].set_value(
            0 # Default value
        )
        # User defined: Fraction of total condensate that returns to boiler (dimensionless)
        self.cond_return_rate = Var(
            self.flowsheet().time, 
            domain=NonNegativeReals,
            bounds=(0,1),
            doc="Fraction of condensate returned to the boiler. Default: 0.7."
        )
        self.cond_return_rate.fix(
            0.7 # Default value
        )
        # User defined: Condensate return temperature (degC or K)
        self.cond_return_temperature = Var(
            self.flowsheet().time, 
            domain=NonNegativeReals,
            doc="Temperature at which the condensate returns to the boiler. Default: 80 degC.",
            units=units_meta.get_derived_units("temperature"),
        )
        self.cond_return_temperature.fix(
            (80 + 273.15) * UNIT.K # Default fixed value
        )
        # User defined: Subcooling target delta T for condensate after process heating (K)
        self.deltaT_subcool = Var(
            self.flowsheet().time, 
            domain=NonNegativeReals,
            doc="The target amount of subcooling of the condensate after process heating. Default: 0 K.",
            units=units_meta.get_derived_units("temperature"),
        )
        self.deltaT_subcool.fix(
            0 # Default fixed value
        )
        # User defined: Heat and pressure losses
        self.heat_loss = Var(
            self.flowsheet().time,
            domain=NonNegativeReals,
            doc="Heat loss. Default: 0 kW.",
            units=units_meta.get_derived_units("power")
        )
        self.heat_loss.fix(
            0 # Default fixed value
        )        
        self.pressure_loss = Var(
            self.flowsheet().time,
            domain=NonNegativeReals,
            doc="Pressure loss. Default: 0 Pa.",
            units=units_meta.get_derived_units("pressure")
        )
        self.pressure_loss.fix(
            0 # Default fixed value
        )   
        
        # User defined when desuperheating is active, otherwise do not show
        if self.config.has_desuperheating:
            # User defined: Boiler feed water temperature entering the desuperheater
            self.bfw_temperature = Var(
                self.flowsheet().time, 
                domain=NonNegativeReals,
                doc="The target amount of subcooling of the condensate after process heating. Default: 0 K.",
                units=units_meta.get_derived_units("temperature"),
            )
            self.bfw_temperature.fix(
                (110 + 273.15) * UNIT.K # Default fixed value
            )
            # User defined: Target degree of superheat at the outlet of the desuperheater
            self.deltaT_superheat = Var(
                self.flowsheet().time, 
                domain=NonNegativeReals,
                doc="The target amount of superheat present in the steam after desuperheating before use. Default: 0 K.",
                units=units_meta.get_derived_units("temperature"),
            )
            self.deltaT_superheat.fix(
                0 # Default fixed value
            )

    def _create_expressions(self) -> None:
        """Create helper Expressions.

        Creates:
            - ``energy_lost`` 
        """
        # Calculated, always show
        self.energy_lost = Expression(
            self.flowsheet().time,
            rule=lambda b, t: (
                b._int_outlet_cond_state[t].flow_mol
                * (b._int_outlet_cond_state[t].enth_mol - b._ref_enth)
                - 
                b.outlet_return_state[t].flow_mol
                * (b.outlet_return_state[t].enth_mol   - b._ref_enth)
            ),
            doc="Energy lost from condensate cooling and condensate to drain.",
        )

    # ------------------------------------------------------------------
    # Balances
    # ------------------------------------------------------------------
    def _add_material_balances(self) -> None:
        """Material balance equations summary.

        Balances / Constraints:
            - ``overall_material_balance``
            - ``condensate_return_material_eq``
            - ``intermediate_material_balance_post_heating``
            - ``intermediate_material_balance_pre_heating``
        """
        @self.Constraint(
                self.flowsheet().time, 
                doc="Overall material balance",
                )
        def overall_material_balance(b, t):
            return (
                sum(
                    o[t].flow_mol
                    for o in b.outlet_blocks
                )
                == 
                sum(
                    i[t].flow_mol
                    for i in b.inlet_blocks
                )
            )
        @self.Constraint(
                self.flowsheet().time, 
                doc="Intermediate material balance",
                )
        def intermediate_material_balance_pre_heating(b, t):
            return (
                b._int_mixed_inlet_state[t].flow_mol
                ==
                sum(
                    i[t].flow_mol
                    for i in b.inlet_blocks
                ) 
            )             
        @self.Constraint(
                self.flowsheet().time, 
                doc="Intermediate material balance",
                )
        def intermediate_material_balance_post_heating(b, t):
            return (
                b._int_outlet_cond_state[t].flow_mol
                ==
                sum(
                    i[t].flow_mol
                    for i in b.inlet_blocks
                ) 
            )        
        @self.Constraint(
                self.flowsheet().time, 
                doc="Condensate return material equation",
                )
        def condensate_return_material_eq(b, t):
            return (
                b.outlet_return_state[t].flow_mol 
                ==
                sum(
                    i[t].flow_mol
                    for i in b.inlet_blocks
                ) 
                *
                b.cond_return_rate[t]
            )
        
    def _add_energy_balances(self) -> None:
        """Energy balance equations summary.

        Balances / Constraints:
            - ``mixing_energy_balance``
            - ``heating_energy_balance``
        """
        @self.Constraint(
                self.flowsheet().time, 
                doc="Inlet mixing energy balance",
                )
        def mixing_energy_balance(b, t):
            return (
                b._int_mixed_inlet_state[t].flow_mol * b._int_mixed_inlet_state[t].enth_mol
                ==
                sum(
                    i[t].flow_mol * i[t].enth_mol
                    for i in b.inlet_blocks
                )
            )
        @self.Constraint(
                self.flowsheet().time,
                doc="Process heating energy balance",
                )
        def heating_energy_balance(b, t):
            return (
                b._int_mixed_inlet_state[t].flow_mol * b._int_mixed_inlet_state[t].enth_mol
                == 
                b._int_outlet_cond_state[t].flow_mol * b._int_outlet_cond_state[t].enth_mol
                +
                b.heat_loss[t]
                +
                b.heat_demand[t]
            )
        self.outlet_drain_state[:].enth_mol.fix( 
            value(
                self._ref_enth
            )
        )
        @self.Constraint(
                self.flowsheet().time, 
                doc="Saturated liquid enthalpy",
                )            
        def saturated_liq_enthalpy_eq(b, t):
            return (
                b._int_outlet_sat_liq_state[t].enth_mol_sat_phase["Liq"]
                ==
                b._int_outlet_sat_liq_state[t].enth_mol
            )
        if self.config.has_desuperheating:
            @self.Constraint(
                    self.flowsheet().time, 
                    doc="Saturated vapour enthalpy",
                    )            
            def saturated_vap_enthalpy_eq(b, t):
                return (
                    b._int_inlet_sat_vap_state[t].enth_mol_sat_phase["Vap"]
                    ==
                    b._int_inlet_sat_vap_state[t].enth_mol
                )           

    def _add_momentum_balances(self) -> None:
        """Momentum balance equations summary.

        Balances / Constraints:
            - ``mixing_momentum_balance``
            - ``heating_momentum_balance``
            - ``outlet_momentum_balance``
            If desuperheating:
                - ``intlet_water_momentum_balance``            
        """
        @self.Constraint(
                self.flowsheet().time,
                doc="Momentum equalities",
                )
        def mixing_momentum_balance(b, t):
            return (
                b.inlet_steam_state[t].pressure
                ==
                b._int_mixed_inlet_state[t].pressure 
            )
        @self.Constraint(
                self.flowsheet().time, 
                doc="Process heating momentum balance",
                )
        def heating_momentum_balance(b, t):
            return (
                b._int_mixed_inlet_state[t].pressure
                == 
                b._int_outlet_cond_state[t].pressure 
                + 
                b.pressure_loss[t]
            )
        @self.Constraint(
                self.flowsheet().time, 
                doc="Saturated liq pressure",
                )
        def saturated_liq_pressure_eq(b, t):
            return (
                b._int_outlet_sat_liq_state[t].pressure
                == 
                b._int_outlet_cond_state[t].pressure 
            )        
        @self.Constraint(
                self.flowsheet().time, 
                doc="Outlet momentum equality",
                )
        def outlet_momentum_balance(b, t):
            return (
                b.outlet_return_state[t].pressure
                == 
                b.outlet_drain_state[t].pressure
            )
        self.outlet_return_state[:].pressure.fix(
            101325 * UNIT.Pa # Fixed value, hidden from the user
        )
        if self.config.has_desuperheating:
            @self.Constraint(
                    self.flowsheet().time, 
                    doc="Inlet water momentum balance",
                    )
            def intlet_water_momentum_balance(b, t):
                return (
                    b.inlet_water_state[t].pressure
                    == 
                    b.inlet_steam_state[t].pressure 
                )
            @self.Constraint(
                    self.flowsheet().time, 
                    doc="Saturated vapour pressure",
                    )            
            def saturated_vap_pressure_eq(b, t):
                return (
                    b.inlet_steam_state[t].pressure
                    ==
                    b._int_inlet_sat_vap_state[t].pressure
                )                 

    def _add_additional_constraints(self) -> None:
        """Add auxiliary constraints and bounds.

        Constraints:
            - ``condensate_temperature_eq``
            - ``subcooling_temperature_eq``
            If desuperheating:
                - ``desuperheating_mixed_temperature_eq``    
        """
        @self.Constraint(
                self.flowsheet().time, 
                doc="Condensate return temperature",
                )
        def condensate_temperature_eq(b, t):
            return (
                b.outlet_return_state[t].temperature
                == 
                b.cond_return_temperature[t]
            )
        @self.Constraint(
                self.flowsheet().time, 
                doc="Subcool temperature",
                )
        def subcooling_temperature_eq(b, t):
            return (
                b._int_outlet_cond_state[t].temperature
                == 
                b._int_outlet_sat_liq_state[t].temperature - b.deltaT_subcool[t]
            )
        if self.config.has_desuperheating:
            @self.Constraint(
                    self.flowsheet().time, 
                    doc="Inlet water temperature",
                    )
            def inlet_water_temperature_eq(b, t):
                return (
                    b.inlet_water_state[t].temperature
                    == 
                    b.bfw_temperature[t]
                )            
            @self.Constraint(
                    self.flowsheet().time, 
                    doc="Mixed temperature after desuperheating",
                    )
            def desuperheating_mixed_temperature_eq(b, t):
                return (
                    b._int_mixed_inlet_state[t].temperature 
                    ==
                    b._int_inlet_sat_vap_state[t].temperature + b.deltaT_superheat[t]
                )         

    def calculate_scaling_factors(self):
        """Assign scaling factors to improve numerical conditioning.

        Sets scaling factors for performance and auxiliary variables. 
        """
        super().calculate_scaling_factors()
        scaling.set_scaling_factor(self.heat_loss, 1e-6) # kW scale
        scaling.set_scaling_factor(self.pressure_loss, 1e-6) # Pa scale

    def _get_stream_table_contents(self, time_point=0):
        """Create a stream table for all inlets and outlets.

        Args:
            time_point (int | float): Time index at which to extract stream data.

        Returns:
            pandas.DataFrame: A tabular view suitable for reporting via
            ``create_stream_table_dataframe``.
        """
        io_dict = {}

        for inlet_name in self.inlet_list:
            io_dict[inlet_name] = getattr(self, inlet_name)

        for outlet_name in self.outlet_list:
            io_dict[outlet_name] = getattr(self, outlet_name)

        return create_stream_table_dataframe(io_dict, time_point=time_point)

    def _get_performance_contents(self, time_point=0):
        """Collect performance variables for reporting.

        Args:
            time_point (int | float): Time index at which to report values.

        Returns:
            dict: Mapping used by IDAES reporters, containing human-friendly labels
            to Vars/References (e.g., heat/pressure loss, mixed-state properties).
        """
        perf = {
            "vars": {
                "Heat demand [W]": self.heat_demand[time_point],
                "Degree of subcooling target [K]": self.deltaT_subcool[time_point],
                "Heat loss [W]": self.heat_loss[time_point],
                "Pressure loss [Pa]": self.pressure_loss[time_point],
                "Condensate return rate [-]:": self.cond_return_rate[time_point],
                "Condensate return temperature [degC]": UNIT.convert_temp_K_to_C(self.cond_return_temperature[time_point]),
            },
            "exprs": {
                "Energy lost to return network [W]": self.energy_lost[time_point],
            },
        }
        if self.config.has_desuperheating:
            perf["vars"].update(
                {
                    "BFW temperature [K]": self.bfw_temperature[time_point],
                    "Degree of superheat target [K]": self.deltaT_superheat[time_point],
                }
            )
        return perf

    def initialize(self, *args, **kwargs):
        """Initialize the SteamUser unit using :class:`SteamUserInitializer`.

        Args:
            *args: Forwarded to ``SteamUserInitializer.initialize``.
            **kwargs: Forwarded to ``SteamUserInitializer.initialize`` (e.g., solver, options).

        Returns:
            pyomo.opt.results.results_.SolverResults: Results from the initializer's solve.
        """
        init = SteamUserInitializer()
        return init.initialize(self, *args, **kwargs)



### Imports
from pyomo.environ import ConcreteModel, SolverFactory, SolverStatus, TerminationCondition, Block, TransformationFactory, assert_optimal_termination
from pyomo.network import SequentialDecomposition, Port, Arc
from pyomo.core.base.units_container import _PyomoUnit, units as pyomo_units
from idaes.core import FlowsheetBlock
from idaes.core.util.model_statistics import report_statistics, degrees_of_freedom
from idaes.core.util.tables import _get_state_from_port
import idaes.logger as idaeslog
from property_packages.build_package import build_package
from ahuora_builder.custom.custom_compressor import CustomCompressor


### Utility Methods
def units(item: str) -> _PyomoUnit:
    ureg = pyomo_units._pint_registry
    pint_unit = getattr(ureg, item)
    return _PyomoUnit(pint_unit, ureg)


### Build Model
m = ConcreteModel()
m.fs = FlowsheetBlock(dynamic=False)

# Set up property packages
m.fs.PP__1 = build_package(
    "helmholtz",
    ["h2o"],
)

# Create unit models

# Compressor-1
m.fs.Compressor_1 = CustomCompressor(
    property_package=m.fs.PP__1,
    power_property_package=m.fs.power_property_package,
    dynamic=False
)
m.fs.Compressor_1.deltaP.fix(50000.0 * units("Pa"))
m.fs.Compressor_1.efficiency_isentropic.fix(0.9 * units("None"))
sb = _get_state_from_port(m.fs.Compressor_1.inlet, 0)
sb.constrain_component(sb.pressure, 101325.0 * units("Pa"))
sb.constrain_component(sb.enth_mol, 4000.0 * units("J/mol"))
sb.constrain_component(sb.flow_mol, 100.0 * units("mol/s"))
sb.constrain_component(sb.mole_frac_comp[('h2o',)], 1.0 * units("None"))


### Check Model Status
report_statistics(m)
print("Degrees of freedom:", degrees_of_freedom(m))


### Initialize Model
m.fs.Compressor_1.initialize(outlvl=idaeslog.INFO)


### Solve
opt = SolverFactory("ipopt")
res = opt.solve(m, tee=True)
assert_optimal_termination(res)


### Report
m.fs.Compressor_1.report()



from pyomo.core.base.indexed_component import UnindexedComponent_set, IndexedComponent



def is_scalar_reference(component: IndexedComponent) -> bool:
    # The only key in it should be None
    return list(component) == [None]


# Import Pyomo libraries
from stringprep import in_table_a1
from pyomo.environ import (
    Constraint,
    Set,
    Var,
    Suffix,
    units as pyunits,
)
from pyomo.environ import Reals
from pyomo.common.config import ConfigBlock, ConfigValue, In
from idaes.core.util.model_statistics import degrees_of_freedom
# Import IDAES cores
from idaes.core import (
    declare_process_block_class,
    UnitModelBlockData,
    useDefault,
    MaterialBalanceType,
    MaterialFlowBasis,
)
from idaes.core.util.config import is_physical_parameter_block
import idaes.core.util.scaling as iscale
import idaes.logger as idaeslog
from idaes.models.unit_models import separator
from idaes.core.util.tables import create_stream_table_dataframe
from idaes.core.util.exceptions import ConfigurationError, BurntToast, PropertyNotSupportedError

from idaes.models.unit_models import separator
#Import enum
from enum import Enum

# Set up logger
_log = idaeslog.getLogger(__name__)

# Enumerate options for balances
class SplittingType(Enum):
    """
    Enum of supported material split types.
    """

    totalFlow = 1
    phaseFlow = 2
    componentFlow = 3
    phaseComponentFlow = 4



# When using this file the name "acBus" is what is imported
@declare_process_block_class("acBus")
class acBusData(UnitModelBlockData):
    """
    Zero order acbus model
    """

    # CONFIG are options for the unit model, this simple model only has the mandatory config options
    CONFIG = ConfigBlock()

    CONFIG.declare(
        "dynamic",
        ConfigValue(
            domain=In([False]),
            default=False,
            description="Dynamic model flag - must be False",
            doc="""Indicates whether this model will be dynamic or not,
    **default** = False. The Bus unit does not support dynamic
    behavior, thus this must be False.""",
        ),
    )
    CONFIG.declare(
        "has_holdup",
        ConfigValue(
            default=False,
            domain=In([False]),
            description="Holdup construction flag - must be False",
            doc="""Indicates whether holdup terms should be constructed or not.
    **default** - False. The Bus unit does not have defined volume, thus
    this must be False.""",
        ),
    )
    CONFIG.declare(
        "property_package",
        ConfigValue(
            default=useDefault,
            domain=is_physical_parameter_block,
            description="Property package to use for control volume",
            doc="""Property parameter object used to define property calculations,
    **default** - useDefault.
    **Valid values:** {
    **useDefault** - use default package from parent model or flowsheet,
    **PhysicalParameterObject** - a PhysicalParameterBlock object.}""",
        ),
    )
    CONFIG.declare(
        "property_package_args",
        ConfigBlock(
            implicit=True,
            description="Arguments to use for constructing property packages",
            doc="""A ConfigBlock with arguments to be passed to a property block(s)
    and used when constructing these,
    **default** - None.
    **Valid values:** {
    see property package for documentation.}""",
        ),
    )
    CONFIG.declare(
        "num_inlets",
        ConfigValue(
            default=False,
            domain=int,
            description="Number of inlets to add",
            doc="Number of inlets to add",
        ),
    )
    CONFIG.declare(
        "num_outlets",
        ConfigValue(
            default=False,
            domain=int,
            description="Number of outlets to add",
            doc="Number of outlets to add",
        ),
    )
    CONFIG.declare(
        "material_balance_type",
        ConfigValue(
            default=MaterialBalanceType.useDefault,
            domain=In(MaterialBalanceType),
            description="Material balance construction flag",
            doc="""Indicates what type of mass balance should be constructed,
            **default** - MaterialBalanceType.useDefault.
            **Valid values:** {
            **MaterialBalanceType.useDefault - refer to property package for default
            balance type
            **MaterialBalanceType.none** - exclude material balances,
            **MaterialBalanceType.componentPhase** - use phase component balances,
            **MaterialBalanceType.componentTotal** - use total component balances,
            **MaterialBalanceType.elementTotal** - use total element balances,
            **MaterialBalanceType.total** - use total material balance.}""",
        ),
    )


    def build(self):
        # build always starts by calling super().build()
        # This triggers a lot of boilerplate in the background for you
        super().build()

        # This creates blank scaling factors, which are populated later
        self.scaling_factor = Suffix(direction=Suffix.EXPORT)


        # Defining parameters of state block class
        tmp_dict = dict(**self.config.property_package_args)
        tmp_dict["parameters"] = self.config.property_package
        tmp_dict["defined_state"] = True  # inlet block is an inlet

        # Add state blocks for inlet, outlet, and waste
        # These include the state variables and any other properties on demand
        num_inlets = self.config.num_inlets
        self.inlet_list = [ "inlet_" + str(i+1) for i in range(num_inlets) ]    
        self.inlet_set = Set(initialize=self.inlet_list) 
        self.inlet_blocks = []

        for name in self.inlet_list:
            # add properties_inlet_1, properties_inlet2 etc
            state_block = self.config.property_package.state_block_class(
                self.flowsheet().config.time, doc="inlet power", **tmp_dict
            )
            self.inlet_blocks.append(state_block)
            # Dynamic equivalent to self.properties_inlet_1 = stateblock
            setattr(self,"properties_" + name, state_block)
            # also add the port
            self.add_port(name=name,block=state_block)


        # Add outlet state blocks

        num_outlets = self.config.num_outlets
        self.outlet_list = [ "outlet_" + str(i+1) for i in range(num_outlets) ]
        self.outlet_set = Set(initialize=self.outlet_list)
        self.outlet_blocks = []

        for name in self.outlet_list:
            tmp_dict["defined_state"] = False
            state_block = self.config.property_package.state_block_class(
                self.flowsheet().config.time, doc="outlet power", **tmp_dict
            )
            self.outlet_blocks.append(state_block)
            setattr(self,"properties_" + name, state_block)
            self.add_port(name=name,block=state_block)
 
        # Add variable for power splitting
        self.split_fraction = Var(
            self.flowsheet().time,
            self.outlet_set,
            initialize=1.0,
            #units = pyunits.dimensionless,
            doc="How the power is split between outlets",
        )
       
        #Obtain the power components from the inlets
        @self.Expression(
            self.flowsheet().time,
        )
        def total_active_power(b,t):
            return  sum(state_block[t].active_power for state_block in self.inlet_blocks)
        
        @self.Expression(
            self.flowsheet().time,
        )
        def total_reactive_power(b,t):
            return  sum(state_block[t].reactive_power for state_block in self.inlet_blocks)
        
        @self.Expression(
                self.flowsheet().time,
        )
        def total_voltage(b,t):
            return  sum(state_block[t].voltage for state_block in self.inlet_blocks)

        #Add constraints
    
        @self.Constraint(
            self.flowsheet().time,
            self.outlet_list,
            doc="active power split",   
        )
        def eq_active_power(b,t,o):
            outlet_block = getattr(self,"properties_" + o)
            return outlet_block[t].active_power == (
                self.total_active_power[t] * b.split_fraction[t,o]) 

        @self.Constraint(
            self.flowsheet().time,
            self.outlet_list,
            doc="reactive power split", 
        )
        def eq_reactive_power(b,t,o):
            outlet_block = getattr(self,"properties_" + o)
            return outlet_block[t].reactive_power == (
                self.total_reactive_power[t] * b.split_fraction[t,o])
       
        @self.Constraint(
            self.flowsheet().time,
            self.outlet_list,
            doc="voltage split", 
        )
        def eq_voltage(b,t,o):
            outlet_block = getattr(self,"properties_" + o)
            return outlet_block[t].voltage == (
                self.total_voltage[t] * b.split_fraction[t,o])    


        @self.Constraint(
            self.flowsheet().time,
            doc="Split fraction sum to 1",
        )
        def eq_split_fraction_sum(b, t):
            return sum(b.split_fraction[t, o] for o in self.outlet_list) == 1.0     
        
    
    def calculate_scaling_factors(self):
        super().calculate_scaling_factors()
    
    def initialize(blk, *args, **kwargs):
        
       pass


    def _get_stream_table_contents(self, time_point=0):

        io_dict = {}
        for inlet_name in self.inlet_list:
            io_dict[inlet_name] = getattr(self, inlet_name) # get a reference to the port
        
        out_dict = {}
        for outlet_name in self.outlet_list:
            out_dict[outlet_name] = getattr(self, outlet_name) # get a reference to the port


   



### Imports
from pyomo.environ import ConcreteModel, SolverFactory, SolverStatus, TerminationCondition, Block, TransformationFactory, assert_optimal_termination
from pyomo.network import SequentialDecomposition, Port, Arc
from pyomo.core.base.units_container import _PyomoUnit, units as pyomo_units
from idaes.core import FlowsheetBlock
from idaes.core.util.model_statistics import report_statistics, degrees_of_freedom
from idaes.core.util.tables import _get_state_from_port
import idaes.logger as idaeslog
from property_packages.build_package import build_package
from ahuora_builder.custom.custom_pump import CustomPump
from ahuora_builder.custom.custom_heater import DynamicHeater


### Utility Methods
def units(item: str) -> _PyomoUnit:
    ureg = pyomo_units._pint_registry
    pint_unit = getattr(ureg, item)
    return _PyomoUnit(pint_unit, ureg)

def init_unit(unit: Block) -> None:
    unit.initialize(outlvl=idaeslog.INFO)


### Build Model
m = ConcreteModel()
m.fs = FlowsheetBlock(dynamic=False)

# Set up property packages
m.fs.PP__1 = build_package(
    "helmholtz",
    ["h2o"],
)

# Create unit models

# Pump-1
m.fs.Pump_1 = CustomPump(
    property_package=m.fs.PP__1,
    dynamic=False,
    power_property_package=m.fs.power_property_package
)
m.fs.Pump_1.efficiency_pump.fix(0.8 * units("dimensionless"))
sb = _get_state_from_port(m.fs.Pump_1.inlet, 0)
sb.constrain_component(sb.pressure, 11000000.0 * units("Pa"))
sb.constrain_component(sb.enth_mol, 125.0 * units("J/mol"))
sb.constrain_component(sb.flow_mol, 431.0 * units("mol/s"))
sb = _get_state_from_port(m.fs.Pump_1.outlet, 0)
sb.constrain_component(sb.pressure, 22000000.0 * units("Pa"))

# Heater-1
m.fs.Heater_1 = DynamicHeater(
    property_package=m.fs.PP__1,
    has_pressure_change=None,
    dynamic=None,
    has_holdup=None
)
m.fs.Heater_1.heat_duty.fix(100.0 * units("J/s"))


### Connect Unit Models
m.fs.arc_1 = Arc(source=m.fs.Pump_1.outlet, destination=m.fs.Heater_1.inlet)


### Check Model Status
report_statistics(m)
print("Degrees of freedom:", degrees_of_freedom(m))


### Initialize Model
TransformationFactory("network.expand_arcs").apply_to(m)
seq = SequentialDecomposition()
seq.set_tear_set([])
seq.run(m, init_unit)


### Solve
opt = SolverFactory("ipopt")
res = opt.solve(m, tee=True)
assert_optimal_termination(res)


### Report
m.fs.Pump_1.report()
m.fs.Heater_1.report()



from ..direct_steam_injection import Dsi
import pytest
import pyomo.environ as pyo
from pyomo.network import Arc
from idaes.core import FlowsheetBlock, MaterialBalanceType
from idaes.models.unit_models import Heater, Valve
from idaes.models.properties import iapws95
from idaes.core.util.initialization import propagate_state
from idaes.core.util.model_statistics import degrees_of_freedom
# need to pip install ahuora_compounds@git+https://github.com/waikato-ahuora-smart-energy-systems/PropertyPackages.git@v0.0.25
from property_packages.build_package import build_package

def test_dsi():
    m = pyo.ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)
    m.fs.steam_properties = build_package("helmholtz", ["water"],["Liq","Vap"])
    m.fs.milk_properties = build_package("milk", ["water","milk_solid"],["Liq","Vap"])
    m.fs.dsi = Dsi(property_package=m.fs.milk_properties,steam_property_package=m.fs.steam_properties)
    
    m.fs.dsi.inlet.flow_mol.fix(1)
    m.fs.dsi.properties_milk_in[0].constrain_component(m.fs.dsi.properties_milk_in[0].temperature, 300 * pyo.units.K)
    m.fs.dsi.inlet.pressure.fix(101325)
    m.fs.dsi.inlet.mole_frac_comp[0,"water"].fix(0.9)
    m.fs.dsi.inlet.mole_frac_comp[0,"milk_solid"].fix(0.1)

    m.fs.dsi.steam_inlet.flow_mol.fix(1)
    m.fs.dsi.properties_steam_in[0].constrain_component(m.fs.dsi.properties_steam_in[0].temperature, 400 * pyo.units.K)
    m.fs.dsi.steam_inlet.pressure.fix(101325)

    m.fs.dsi.initialize()
    assert degrees_of_freedom(m.fs.dsi.properties_milk_in) == 0
    assert degrees_of_freedom(m.fs.dsi.properties_steam_in) == 0
    assert degrees_of_freedom(m.fs) == 0

    opt = pyo.SolverFactory("ipopt")
    results = opt.solve(m, tee=True)
    
    assert results.solver.termination_condition == pyo.TerminationCondition.optimal
    assert degrees_of_freedom(m.fs) == 0


